/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <string.h>

#include <glib.h>
#include <gnome.h>

#include "acache.h"
#include "article.h"
#include "articlelist.h"
#include "debug.h"
#include "gnksa.h"
#include "server.h"
#include "group.h"
#include "prefs.h"
#include "util.h"

/***
****
***/

static gchar* decode_quoted_printable (const gchar * encode);



/***
****
****  LIFE CYCLE
****
***/

void
article_destructor (PanObject* o)
{
	Article * a = ARTICLE(o);
        g_return_if_fail (a != NULL);

	/* clean up article parts */
	debug1 (DEBUG_PAN_OBJECT, "article_destructor: %p", a);
	g_slist_free (a->threads);
	if (a->headers != NULL)
		g_hash_table_destroy (a->headers);

	/* clean up parent */
	pan_object_destructor (o);
}

void
article_constructor (Article * article, Group * group)
{
	g_return_if_fail (article!=NULL);
	g_return_if_fail (group!=NULL);

	/* init the superclass */
	pan_object_constructor (PAN_OBJECT(article), article_destructor);

	/* init this class */
        debug1 (DEBUG_PAN_OBJECT, "article_constructor: %p", article);
	article->group = group;
	article->self_passes_filter = FALSE;
	article->tree_passes_filter = FALSE;
	article->unread_children = 0;
	article->parts = 0;
	article->part = 0;
	article->linecount = 0;
	article->crosspost_qty = 0;
	article->state = 0;
	article->date = (time_t)0;
	article->number = 0;
	article->subject = NULL;
	article->author_real = NULL;
	article->author_addr = NULL;
	article->message_id = NULL;
	article->references = NULL;
	article->parent = NULL;
	article->threads = NULL;
	article->headers = NULL;
}

Article*
article_new (Group * group)
{
	Article* article = g_new (Article, 1);
	article_constructor (article, group);
	return article;
}


static void
article_dup_ghfunc (gpointer key, gpointer val, gpointer user_data)
{
	Article * article = ARTICLE(user_data);
	article_set_header_pair_chunk (article,
		group_chunk_string (article->group, (const gchar*)key),
		group_chunk_string (article->group, (const gchar*)val));
}
Article*
article_dup (Group * group, Article * article)
{
	Article * retval;

	g_return_val_if_fail (group!=NULL, NULL);
	g_return_val_if_fail (article!=NULL, NULL);

#define chunk_if_exists(A) \
	is_nonempty_string (A) ? group_chunk_string (group, A) : NULL

	retval                      = article_new (group);
	retval->self_passes_filter  = FALSE;
	retval->tree_passes_filter  = FALSE;
	retval->unread_children     = 0;
	retval->group               = group;
	retval->parts               = article->parts;
	retval->part                = article->part;
	retval->linecount           = article->linecount;
	retval->crosspost_qty       = article->crosspost_qty;
	retval->state               = article->state;
	retval->date                = article->date;
	retval->number              = article->number;
	retval->subject             = chunk_if_exists (article->subject);
	retval->author_real         = chunk_if_exists (article->author_real);
	retval->author_addr         = chunk_if_exists (article->author_addr);
	retval->message_id          = chunk_if_exists (article->message_id);
	retval->references          = chunk_if_exists (article->references);
	retval->parent              = NULL;
	retval->threads             = NULL;
	retval->headers             = NULL;

	if (article->headers != NULL)
		g_hash_table_foreach (article->headers, article_dup_ghfunc, retval);

	return retval;
}

/***
****
***/

static void
article_set_dirty (Article * article)
{
	g_return_if_fail (article != NULL);
	g_return_if_fail (article->group != NULL);

	article->group->articles_dirty = TRUE;
	articlelist_update_node (article);
}



/**
 * Remove an article from its parent/child threadings.  The children and
 * parent are linked together instead.  This is useful for not breaking
 * the threading of articles surrounding an article which is about to be
 * destroyed.
 */
void
article_isolate (Article* article)
{
	GSList *l = NULL;
	Article *parent = article->parent;

	if (parent)
		parent->threads = g_slist_remove (parent->threads, article);

	for (l=article->threads; l!=NULL; l=l->next) {
		Article *child = ARTICLE(l->data);
		child->parent = parent;
		if (parent)
			parent->threads = g_slist_append (parent->threads, child);
	}

	g_slist_free (article->threads);
	article->threads = NULL;
	article->parent = NULL;
}


gboolean
article_is_new (const Article * a)
{
	g_return_val_if_fail (a!=NULL, FALSE);
	g_return_val_if_fail (a->group!=NULL, FALSE);
	return a->number > a->group->article_high_old;
}

/***
****
****  READ / UNREAD
****
***/

gboolean
article_is_read (const Article * a)
{
	g_return_val_if_fail (a!=NULL, FALSE);
	g_return_val_if_fail (a->group!=NULL, FALSE);

	return newsrc_is_article_read (a->group->newsrc, a->number);
}

/**
 * Mark the article (and possibly multipart children) unread.
 * @param article the article to mark as unread.
 * @return the number of articles marked unread in this call to this function.
 */
static int
article_set_unread_impl (Article * article)
{
	int unread_qty = 0;
	const gboolean was_read = newsrc_mark_article (article->group->newsrc,
	                                               article->number,
	                                               FALSE);
	if (was_read) {
		++unread_qty;
		articlelist_article_read_changed (article);
	}

	/* FIXME: I don't think this special case is such a great idea.
	   Special case: if it's a multipart, mark 1..n as read too */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				unread_qty += article_set_unread_impl (child);
		}
	}

	return unread_qty;
}

/**
 * Mark the article (and possibly multipart children) read.
 * @param article the article to mark as read.
 * @return the number of articles marked read in this call to this function.
 */
static int
article_set_read_impl (Article * article)
{
	int read_qty = 0;
	const gboolean was_read = newsrc_mark_article (article->group->newsrc,
	                                               article->number,
	                                               TRUE);
	if (!was_read) {
		++read_qty;
		articlelist_article_read_changed (article);
	}

	/* FIXME: I don't think this special case is such a great idea.
	   Special case: if it's a multipart, mark 1..n as read too */
	if (article->part>0 && article->parts>1 && article->threads!=NULL) {
		GSList * l;
		for (l=article->threads; l!=NULL; l=l->next) {
			Article * child = ARTICLE(l->data);
			if (child->part > 0)
				read_qty += article_set_read_impl (child);
		}
	}

	return read_qty;
}

void
articles_set_read (Article ** articles, int article_qty, gboolean read)
{
	int i;
	int read_inc;

	g_return_if_fail (articles != NULL);

	for (read_inc=i=0; i<article_qty; ++i)
	{
		Article * article = articles[i];

		g_return_if_fail (article != NULL);
		g_return_if_fail (article->group != NULL);
		g_assert (articles[0]->group == article->group);

		if (read)
			read_inc += article_set_read_impl (article);
		else 
			read_inc -= article_set_unread_impl (article);
	}

	if (read_inc)
		group_inc_article_read_qty (articles[0]->group, read_inc);
}

void
article_set_read (Article * article, gboolean read)
{
	articles_set_read (&article, 1, read);
}
        
        


/**
*** PUBLIC MUTATORS
**/

void
article_set_string_field (Article       * article,
                          const gchar  ** field_ptr,
			  const gchar   * chunkme_replacement)
{
	g_return_if_fail (article != NULL);
	g_return_if_fail (article->group != NULL);
	g_return_if_fail (field_ptr != NULL);

	*field_ptr = group_chunk_string (article->group, chunkme_replacement);
	article_set_dirty (article);
}

void
article_set_encoding_chunk (Article         * article,
                            const gchar     * encoding)
{
	g_return_if_fail (article != NULL);
	article_set_header_pair_chunk (article, HEADER_ENCODING, encoding);
}

void
article_set_reply_to_chunk (Article         * article,
	                    const gchar     * reply_to)
{
	g_return_if_fail (article != NULL);
	article_set_header_pair_chunk (article, HEADER_REPLY_TO, reply_to);
}

void
article_set_followup_to_chunk (Article         * article,
                               const gchar     * followup_to)
{
	g_return_if_fail (article != NULL);
	article_set_header_pair_chunk (article, HEADER_FOLLOWUP_TO, followup_to);
}

void
article_set_newsgroups_chunk (Article     * article,
	                      const gchar * newsgroups)
{
	g_return_if_fail (article != NULL);
	article_set_header_pair_chunk (article, HEADER_NEWSGROUPS, newsgroups);
}

void
article_add_flag (Article  * article,
                  gushort    flag)
{
	g_return_if_fail (article != NULL);
	article->state |= flag;
	article_set_dirty (article);
}


void
article_remove_flag (Article         * article,
                     gushort           flag)
{
	g_return_if_fail (article != NULL);
	article->state &= ~flag;
	article_set_dirty (article);
}

void
article_set_flag (Article * a, gushort flag, gboolean b)
{
	g_return_if_fail (a != NULL);
	if (b)
		a->state |= flag;
	else
		a->state &= ~flag;
	article_set_dirty (a);
}


/***
****
****  THREADS
****
***/

gchar*
article_get_thread_message_id (const Article* article)
{
	char * pch;

	g_return_val_if_fail (article!=NULL, NULL);
	g_return_val_if_fail (is_nonempty_string(article->message_id), NULL);

	if (is_nonempty_string(article->references))
	{
		pch = get_next_token_str (article->references, ' ', NULL);
	}
	else /* top of the thread */
	{
		pch = g_strdup (article->message_id);
	}

	return pch;
}

static void
article_get_entire_thread_impl (Article* top, GPtrArray* setme)
{
	GSList *l;
	g_ptr_array_add (setme, top);
	for (l=top->threads; l!=NULL; l=l->next)
		article_get_entire_thread_impl (ARTICLE(l->data), setme);
}

static void
article_get_thread_impl (Article * article,
                         GPtrArray * setme,
                         gboolean children_only)
{
	Article* top = article;

	if (top != NULL)
	{
		if (!children_only)
			while (top->parent != NULL)
				top = top->parent;

		pan_g_ptr_array_reserve (setme, 128);
		article_get_entire_thread_impl (top, setme);
	}
}

void
article_get_subthread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, TRUE);
}

void
article_get_entire_thread (Article* article, GPtrArray* setme)
{
	article_get_thread_impl (article, setme, FALSE);
}


/***
****
****  OTHER HEADERS
****
***/

void
article_remove_header (Article * a, const gchar * header_name)
{
	g_return_if_fail (a!=NULL);
	g_return_if_fail (header_name!=NULL);

	if (a->headers!=NULL)
	{
		g_hash_table_remove (a->headers, header_name);

		if (!g_hash_table_size(a->headers))
		{
			g_hash_table_destroy (a->headers);
			a->headers = NULL;
		}
	}
}

const gchar*
article_get_header (const Article * a, const gchar * header_name)
{
	const gchar * retval = NULL;

	pan_warn_if_fail (a!=NULL);

	if (a!=NULL && a->headers!=NULL)
		retval = (const gchar*) g_hash_table_lookup (a->headers, header_name);

	return retval;
}

gboolean
article_header_is_extra (const gchar * key)
{
	/* sanity check */
	if (!is_nonempty_string(key)) return FALSE;

	/* pan internals shouldn't be visible */
	if (!g_strncasecmp(key,"Pan-",4)) return FALSE;

	/* other headers that are handled explicitly elsewhere */
	if (!strcmp(key,HEADER_FOLLOWUP_TO)) return FALSE;
	if (!strcmp(key,HEADER_NEWSGROUPS)) return FALSE;
	if (!strcmp(key,HEADER_ORGANIZATION)) return FALSE;
	if (!strcmp(key,HEADER_REPLY_TO)) return FALSE;

	return TRUE;
}

static void
ensure_extra_headers_exists (Article * a)
{
	if (a->headers == NULL)
		a->headers = g_hash_table_new (g_str_hash, g_str_equal);
}

void
article_set_header_pair_chunk (Article      * a,
                               const gchar  * key,
                               const gchar  * val)
{
	g_return_if_fail (is_nonempty_string(key));

	article_remove_header (a, key);

	if (is_nonempty_string (val))
	{
		/* since key and val are chunked and const; also, everything this
		   hash returns is const, so everything's fine.  But the hash API
		   wants gpointer instead of gconstpointer, so we cast away const */
		ensure_extra_headers_exists (a);
		g_hash_table_insert (a->headers, (gpointer)key, (gpointer)val);
		article_set_dirty (a);
	}
}

static void
article_get_all_headers_ghfunc (gpointer key, gpointer value, gpointer user_data)
{
	GPtrArray * a = (GPtrArray*)user_data;
	g_ptr_array_add (a, key);
	g_ptr_array_add (a, value);
}

GPtrArray*
article_get_all_headers (const Article * a)
{
	GPtrArray * retval = g_ptr_array_new ();

	if (a->headers != NULL)
	{
		pan_g_ptr_array_reserve (retval, g_hash_table_size(a->headers)*2);
		g_hash_table_foreach (a->headers, article_get_all_headers_ghfunc, retval);
	}

	return retval;
}

gboolean
article_has_attachment (const Article *a)
{
	const gchar * filename = article_get_header (a, PAN_ATTACH_FILE);
	if (filename != NULL)
		return TRUE;
	else
		return FALSE;
}

/***
****
****  BODY
****
***/

gboolean
article_has_body (const Article * a)
{
	gboolean retval = FALSE;

	/* maybe the article's in a folder so the body
	 * is hidden as a header... */
	if (!retval) {
		const gchar * body = article_get_header (a, PAN_BODY);
		if (body != NULL)
			retval = TRUE;
	}

	/* try to get it from acache... */
	if (!retval && is_nonempty_string(a->message_id))
		retval = acache_body_exists (a->message_id);

	return retval;
}

gchar*
article_get_author_str (const Article * a)
{
        gchar * retval;

	if (a==NULL)
	{
		retval = g_strdup ("");
	}
	else if (is_nonempty_string(a->author_addr))
	{
		if (is_nonempty_string(a->author_real))
		{
			retval = g_strdup_printf ("\"%s\" <%s>", a->author_real, a->author_addr);
		}
		else
		{
			retval = g_strdup (a->author_addr);
		}
	}
	else if (is_nonempty_string(a->author_real))
	{
		retval = g_strdup (a->author_real);
	}
	else
	{
		retval = g_strdup ("");
	}

	return retval;
}

gchar*
article_get_body (const Article * a)
{
	gchar * retval = NULL;

	if (a != NULL)
	{
		const gchar * encoding = article_get_header (a, HEADER_ENCODING);

		/* maybe the article's in a folder so the body
		 * is hidden as a header... */
		if (retval == NULL) {
			const gchar * body = article_get_header (a, PAN_BODY);
			if (body != NULL)
				retval = g_strdup (body);
		}

		/* try to get it from acache... */
		if (retval==NULL && is_nonempty_string(a->message_id))
			retval = acache_load_body (a->message_id);

		/* if encoding exists and is quoted-printable, decode it */
		if (is_nonempty_string(encoding))
		{
			/* not all agents give this header (OE doesn't) */
			if (!strcmp (encoding, "quoted-printable"))
				replace_gstr (&retval, decode_quoted_printable(retval));
		}
	}
	else pan_warn_if_reached ();

	return retval;
}

static gchar*
decode_quoted_printable (const gchar * encoded)
{
	gchar *decoded, *out;
	const gchar * in;

	g_return_val_if_fail (is_nonempty_string(encoded), NULL);

	/* we know the decoded str will be no longer than encoded,
	   so use that as the memory allocation upper bound */
       	out = decoded = g_new0 (gchar, strlen(encoded)+1);

	/* decode the string */
	for (in=encoded; *in; ++in)
	{
		if (in[0]!='=')
		{
			*out++ = *in;
		}
		else if (in[1]=='\n')
		{
			/* substitute with nothing */
			in += 2;
		}
		else if ((in[1]=='\r') && (in[2]=='\n'))
		{
			/* substitute with nothing */
			in += 2;
		}
		else if ( ((in[1]==' ')&&(in[2]=='\t')) && ( ((in[3]=='\r') && (in[4]=='\n')) || (in[3]=='\n')) )
		{
			*out++ = *in++;
			*out++ = *in;
		}
		else
		{
			char hex[3];

			/* create a hex string */
			hex[0] = in[1];
			hex[1] = in[2];
			hex[2] = '\0';

			/* convert hex to int ascii value */
			*out++ = (char) strtol (hex, NULL, 16);

			/* skip the octet we just processed */
			in += 2;
		}

		/* g_message ("%s", decoded); */
	}

	*out = '\0';
	return decoded;
}

gchar*
article_get_attribution_string (const Article * a)
{
	gchar * pch = NULL;
	gchar * retval = NULL;
	const gchar * cpch = NULL;
	GString * gstr = NULL;

	pan_warn_if_fail (is_nonempty_string(attribution_line));

	/* get the unsubstituted string */
	gstr = g_string_new (attribution_line);

	/* substitutions: message-id */
	cpch = a==NULL ? "" : a->message_id;
	if (cpch != NULL)
		pan_g_string_replace (gstr, "%i", cpch);

	/* substitutions: date */
	pch = a==NULL ? g_strdup("") : rfc822_date_generate(a->date);
	pan_g_string_replace (gstr, "%d", cpch);
	g_free (pch);

	/* substitutions: author */
	pch = a==NULL ? g_strdup("") : article_get_author_str (a);
	pan_g_string_replace (gstr, "%a", pch);
	g_free (pch);

	/* cleanup & return the result */
	retval = gstr->str;
	g_string_free (gstr, FALSE);
	return retval;
}

void
article_set_author_from_header (Article * a, const gchar * header_from)
{
	gchar * author_addr = NULL;
	gchar * author_real = NULL;

	g_return_if_fail (a!=NULL);
	g_return_if_fail (is_nonempty_string(header_from));

	/* note that we don't check gnksa's return value here,
	   because often posted messages on usenet are 
	   syntactically invalid */
	gnksa_do_check_from (header_from, &author_addr, &author_real);
	if (author_addr != NULL) {
		article_set_string_field (a, &a->author_addr, author_addr);
	}
	if (is_nonempty_string(author_real)) {
		gnksa_strip_realname (author_real);
		article_set_string_field (a, &a->author_real, author_real);
	}

	g_free (author_addr);
	g_free (author_real);
}

