/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <gnome.h>

#include "bozo.h"
#include "gnksa.h"
#include "prefs.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

static GHashTable * hash = NULL;

static gboolean dirty = FALSE;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PRIVATE
************/

/*****
******
*****/

static gchar*
get_bozo_filename (void)
{
	return g_strdup_printf ("%s/bozo.txt", data_dir);
}

/*****
******
*****/

static void
bozo_init_module (void)
{
	if (hash == NULL)
	{
		gchar * filename;
		GArray * file;
		gchar ** v;
		gint i;

		/* create the hashtable */
		hash = g_hash_table_new (g_str_hash, g_str_equal);
		dirty = FALSE;

		/* load the file */
		filename = get_bozo_filename ();
		file = read_file (filename);
		v = file==NULL ? NULL : g_strsplit (file->data, "\n", -1);
		for (i=0; v!=NULL && v[i]!=NULL; ++i)
		{
			gchar * addr = NULL;
			gchar * real = NULL;
			const gchar * line = v[i];

			if (*line=='#')
				continue;

			gnksa_do_check_from (line, &addr, &real);
			if (real==NULL) {
				g_free (addr);
				g_free (real);
				continue;
			}

			gnksa_strip_realname (real);
			g_hash_table_insert (hash, addr, real);
		}

		/* cleanup */
		g_free (filename);
		if (file != NULL)
			g_array_free (file, TRUE);
		g_strfreev (v);
	}
}

/*****
******
*****/

static void
bozo_hash_freekeyval (gpointer key, gpointer value, gpointer user_data)
{
	g_free (key);
	g_free (value);
}
static void
bozo_free_hash (void)
{
	if (hash != NULL)
	{
		g_hash_table_foreach (hash, bozo_hash_freekeyval, NULL);
		g_hash_table_destroy (hash);
		hash = NULL;
	}
}

/*****
******
*****/

static void
bozo_write_line (gpointer key, gpointer value, gpointer user_data)
{
	gchar * address = (gchar*) key;
	gchar * realname = (gchar*) value;
	FILE * fp = (FILE*)user_data;
	fprintf (fp, "\"%s\" <%s>\n", realname, address);
}

/*****
******
*****/


/*****
******
*****/


/************
*************  PUBLIC
************/

gboolean
address_is_bozo (const gchar * address)
{
	gboolean retval = FALSE;

	if (is_nonempty_string(address))
	{
		if (hash == NULL)
			bozo_init_module ();
		retval = g_hash_table_lookup (hash, address) != NULL;
	}

	return retval;
}

void
bozo_clear (void)
{
	if (hash != NULL)
		bozo_free_hash ();
	hash = g_hash_table_new (g_str_hash, g_str_equal);
	dirty = TRUE;
}

void
bozo_remove (const gchar * address)
{
	gpointer orig_key;
	gpointer orig_val;

	g_return_if_fail (is_nonempty_string(address));

	if (hash == NULL)
		bozo_init_module ();

	orig_key = NULL;
	orig_val = NULL;
	if (g_hash_table_lookup_extended (hash, address, &orig_key, &orig_val))
	{
		g_hash_table_remove (hash, orig_key);
		g_free (orig_key);
		g_free (orig_val);
		dirty = TRUE;
	}
}

void
bozo_add (const gchar * address, const gchar * realname)
{
	g_return_if_fail (is_nonempty_string(address));
	g_return_if_fail (is_nonempty_string(realname));
	if (hash == NULL)
		bozo_init_module ();
	g_hash_table_insert (hash, g_strdup(address), g_strdup(realname));
	dirty = TRUE;
}

void
bozo_shutdown_module (void)
{
	if (dirty)
	{
		/* write the bozo file to disk */
		gchar * filename = get_bozo_filename ();
		FILE * fp = fopen (filename, "w+");
		fprintf (fp, "# Bozo file used by the newsreader Pan\n");
		fprintf (fp, "# 1 bozo per line, \"Name\" <mail@domain>\n");
		g_hash_table_foreach (hash, bozo_write_line, fp);
		fclose (fp);
		g_free (filename);
		dirty = FALSE;
	}

	bozo_free_hash ();
}


/*****
******
*****/

static void
bozo_get_list_ghfunc (gpointer key, gpointer val, gpointer user_data)
{
	GPtrArray * a = (GPtrArray*) user_data;
	const gchar * addr = (const gchar*) key;
	const gchar * real = (const gchar*) val;
	gchar * pch;

	if (is_nonempty_string(addr)) {
		if (is_nonempty_string(real)) {
			pch = g_strdup_printf ("\"%s\" <%s>", real, addr);
		} else {
			pch = g_strdup (addr);
		}
	} else if (is_nonempty_string(addr)) {
		pch = g_strdup (real);
	} else {
		pch = g_strdup ("");
	}

	g_ptr_array_add (a, pch);
}

GPtrArray*
bozo_get_list (void)
{
	GPtrArray * a = g_ptr_array_new ();
	if (hash == NULL)
		bozo_init_module ();
	g_hash_table_foreach (hash, bozo_get_list_ghfunc, a);
	return a;

}
