/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-exec.h>
#include <libgnome/gnome-mime.h>
#include <libgnome/gnome-mime-info.h>
#include <libgnome/gnome-i18n.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <pthread.h>

#include "../uulib/uudeview.h"
#include "../uulib/fptools.h"

#include "acache.h"
#include "debug.h"
#include "decode.h"
#include "log.h"
#include "prefs.h" /* for download_dir */
#include "util.h"

pthread_mutex_t uulib_lock = PTHREAD_MUTEX_INITIALIZER;

/**
***  Private Function Prototypes
**/

static char* UUD_file_name_filter (void *opaque, char *fname);
static int   UUD_busy_callback (void *opaque, uuprogress *progress);
static void  UUD_message_callback (void *opaque, char *msg, int level);

static gchar filename_buf[1024];

/**
*** Public Routines
**/

void
decode_open (const char *fname)
{
	const char *mime_type = NULL;
	const char *mime_prog = NULL;
	char **argv;
	int argc;
	
	if (!((mime_type = gnome_mime_type (fname))))
		return;
	if (!((mime_prog = gnome_mime_program (mime_type))))
		return;
	if (!strstr (mime_prog, "%f"))
		return;
	
	argv = g_strsplit (mime_prog, " ", -1);
	for (argc=0; argv[argc]; argc++)
	{
		if (!strcmp(argv[argc], "%f"))
		{
			g_free (argv[argc]);
			argv[argc] = g_strdup (fname);
		}
	}

	gnome_execute_async (NULL, argc, argv);
	g_strfreev (argv);
}

void
decode_article (const decode_data *dd)
{
	GSList *p;
	gint i = 0;
	gint part_qty = 0;
	gint success = 0;
	const gchar * rename_fname = dd->filename;
	Article* article = ARTICLE(dd->articles->data);

	/* sanity checking */
	g_return_if_fail (dd!=NULL);
	g_return_if_fail (dd->server!=NULL);
	g_return_if_fail (dd->articles!=NULL);

	status_item_emit_progress (dd->item, 0);
	status_item_emit_status (dd->item,
				 _("Decoding: %s"), article->subject);

	/* debug... */
	part_qty = g_slist_length(dd->articles);
	debug2 (DEBUG_DECODE,
	        "starting to decode %d-part article \"%s\"",
	        part_qty,
	        article->subject);

	/**
	***  INIT THE DECODE LIBRARY
	**/

	status_item_emit_status (dd->item, _("Waiting to Decode %s"),
				 article->subject);
	pthread_mutex_lock (&uulib_lock);

	status_item_emit_status (dd->item,
				 _("Preparing %s"), article->subject);
	UUInitialize ();

	/* decide where to save the file */
	directory_check (dd->path);
	UUSetOption (UUOPT_SAVEPATH, 0, dd->path);
	UUSetMsgCallback (article->subject, UUD_message_callback);
	UUSetBusyCallback (dd->item, UUD_busy_callback, 1);
	UUSetFNameFilter (NULL, UUD_file_name_filter);

	/**
	***  LOAD THE ARTICLE(S)
	**/

	for (p=dd->articles, i=0; p!=NULL; p=p->next, ++i)
	{
		const Article* article = ARTICLE(p->data);
		int retval;
		gchar* path = g_strdup_printf ("/%s/%s.body",
					       data_dir, article->message_id);
		debug3 (DEBUG_DECODE,
		        "Decoding part %d of %d: %s...", i+1, part_qty, path);
		retval = UULoadFile (path, NULL, 0);
		g_free (path);
		if (retval != UURET_OK)
		{
			/* cleanup */
			UUCleanUp();
			pthread_mutex_unlock (&uulib_lock);
			debug1 (DEBUG_DECODE,
			        "Decode couldn't load tmp file \"%s\"", path);
			return;
		}
	}
	
	status_item_emit_status (dd->item,
				 _("Decoding: %s"), article->subject);

	/**
	***  DECODE
	**/

	for (i=0;; ++i)
	{
		gint val;
		gchar * pch;
		gchar * filename;
		uulist * item;

		/* get the next file from the list */
		item = UUGetFileListItem (i);
		if (item == NULL)
			break;

		/* print some debugging info */
		g_message (
			"part %d, filename %s, subfname %s, "
			"encoding %d, mime id %s, mimetype %s",
			i,
			(item->filename ? item->filename : "(None)"),
			(item->subfname ? item->subfname : "(None)"),
			item->uudet,
			(item->mimeid ? item->mimeid : "(None)"),
			(item->mimetype ? item->mimetype : "(None)"));

		/* check the item's state */
		if (!(item->state & UUFILE_OK))
		{
			pch = g_strdup_printf (
				_("Error parsing part %d of \"%s\": %d"),
				i+1, article->subject, item->state);
			log_add (pch);
			debug0 (DEBUG_DECODE, pch);
			g_free (pch);
			continue;
		}

		/* rename if necessary */
		if (is_nonempty_string (rename_fname)) {
			UURenameFile (item, rename_fname);
			rename_fname = NULL;
		}

		/* try to decode */
		val = UUDecodeFile (item, NULL);
		if (val != UURET_OK) /* no decode */
		{
			pch = g_strdup_printf (
				_("Error decoding part %d of \"%s\": %s"),
				i+1, article->subject, UUstrerror(val));
			log_add (pch);
			debug0 (DEBUG_DECODE, pch);
			g_free (pch);
			continue;
		}

		/* decoded properly, woo-hoo! */
		++success;
		filename = g_strdup(filename_buf);

		/* log the success */
		pch = g_strdup_printf (
			_("Decoded \"%s\" from group \"%s\", \"%s\" part #%d"),
			filename,
			dd->group ? dd->group->name : "Unknown",
			article->subject, i+1);
		log_add (pch);
		debug0 (DEBUG_DECODE, pch);
		g_free (pch);

		/* if this is the first attachment, store the filename */
		if (article_get_header (article, PAN_HEADER_FILENAME) == NULL)
			article_set_header_pair_chunk (article, PAN_HEADER_FILENAME,
				group_chunk_string (article->group, filename));

		/* open as needed... */
		if (dd->open)
			decode_open (filename);

		/* cleanup */
		g_free (filename);
	}

	/**
	***  CLEANUP
	**/

	UUCleanUp ();
	pthread_mutex_unlock (&uulib_lock);

	/* update the node */
	article_remove_flag (article, STATE_DECODE_QUEUED);
	article_add_flag (article, (success?STATE_DECODED:STATE_DECODE_FAILED));
}

/***
**** Private Routines
***/


/***************************************************************
* UUdeview Lib Stuff 
***************************************************************/

static void
UUD_message_callback (void *opaque, char *msg, int level)
{
	debug2 (DEBUG_DECODE,"UUDeview says: \"%s\" (level %d)", msg, level);

	if ((level == UUMSG_WARNING) || (level == UUMSG_ERROR)) {
		pan_error_dialog (_("An error occured during decoding.\n%s"), msg);
	}
}


static int
UUD_busy_callback (void *opaque, uuprogress *progress)
{
	StatusItem *item = STATUS_ITEM(opaque);

	const int i = (progress->percent + 100.0 * (progress->partno-1))
		/ progress->numparts;
	if ( 0<=i && i<=100 )
		status_item_emit_progress (item, i);

	return 0;
}


static char *
UUD_file_name_filter (void *opaque, char *fname)
{
	gchar * buf = filename_buf;
	gchar * pch;

	/* check for empty fname */
	if (fname == NULL)
		return NULL;

	/* copy the filename, including path information */
	if (strchr(fname,'/') == NULL)
	{
		gchar tmp[512];
		UUGetOption (UUOPT_SAVEPATH, NULL, tmp, sizeof(tmp));
		sprintf (buf, "%s/%s", tmp, fname);
	}
	else strcpy (buf, fname);

	/* increment UNKNOWN.??? if necessary */
	pch = strstr (buf, "UNKNOWN.");
	if (pch != NULL)
	{
		pch += 8;
		while (file_exists (buf))
		{
			const gint fnum = atoi(pch) + 1;
			sprintf (pch, "%.3u", fnum);
		}
	}

	/* uniqueness check */
	while (file_exists(buf))
	{
		gint fnum = -1;
		gchar * pch;

		/* find previous number: filename.N */
		pch = strrchr (buf, '.');
		if (pch && isdigit((int)pch[1]))
		{
			gchar * endptr = NULL;
			gint tmpint = strtol (pch+1, &endptr, 10);
			if (!*endptr)
				fnum = tmpint;
		}

		/* update the name */
		if (fnum == -1)
			strcat (buf, ".1");
		else
			sprintf (pch, ".%d", ++fnum);
	}

	/* walk past directory information */
	pch = strrchr (buf, '/');
	if (pch != NULL)
		++pch; /* skip '/' */
	else
		pch = buf;

	return pch;
}
