/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "debug.h"
#include "encode.h"
#include "log.h"
#include "util.h"

/***
 *** encoding table for UUEncode
 ***/

static unsigned char UUEncodeTable[64] = {
    '`', '!', '"', '#', '$', '%', '&', '\'',
    '(', ')', '*', '+', ',', '-', '.', '/',
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '9', ':', ';', '<', '=', '>', '?',
    '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G',
    'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O',
    'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W',
    'X', 'Y', 'Z', '[', '\\',']', '^', '_'
};

/***
 *** Private Function Prototypes
 ***/

static char * encode_part (FILE *stream, int lines_per_part);

/***
 *** Public Routines
 ***/

GPtrArray *
encode_file (const char *filename, const int lines_per_part)
{
	char *part = NULL;
	FILE *fp = NULL;
	GPtrArray *parts = NULL;
	int partno; 
	
	if ((fp = fopen (filename, "r")) == NULL) {
		g_message ("couldnt open file\n");
		return NULL;
	}

	parts = g_ptr_array_new ();
	
	part = g_strdup_printf ("begin 644 %s\n", g_basename(filename));
	g_ptr_array_add (parts, part);
	
	for (partno = 1; (part = encode_part (fp, lines_per_part)); partno++) {
		g_ptr_array_add (parts, part);
	}
	
	part = g_strdup_printf ("%c\nend\n", UUEncodeTable[0]);
	g_ptr_array_add (parts, part);
	
	fclose (fp);
	return parts;
}


/***
**** Private Routines
***/

/* Encode 1 part of the data stream, this was adapted from uulib */

static char *
encode_part (FILE *infile, const int lines_per_part)
{
	unsigned char itemp[256];
	unsigned char otemp[256];
	unsigned char *optr;
	int index, count;
	long line=0;
	GString *part;
	gchar *r_str = NULL;
    
	part = g_string_sized_new (45 * lines_per_part);

	while (!feof (infile) && (lines_per_part <= 0 || line < lines_per_part)) {
		if ((count = fread (itemp, 1, 45, infile)) != 45) {
	    		if (count == 0) {
				break;
			} else if (ferror (infile)) {
				g_message ("Error");
				return NULL;
			}
		}
		optr = otemp;

		/* Encode the number of bytes as first character */
		*optr++ = UUEncodeTable[count];

		for (index=0; index<=count-3; index+=3) {
			*optr++ = UUEncodeTable[itemp[index] >> 2];
			*optr++ = UUEncodeTable[((itemp[index  ] & 0x03) << 4) | (itemp[index+1] >> 4)];
			*optr++ = UUEncodeTable[((itemp[index+1] & 0x0f) << 2) | (itemp[index+2] >> 6)];
			*optr++ = UUEncodeTable[  itemp[index+2] & 0x3f];
		}

		/* Special handling for incomplete lines */
		if (index != count) {
			if (count - index == 2) {
				*optr++ = UUEncodeTable[itemp[index] >> 2];
				*optr++ = UUEncodeTable[((itemp[index  ] & 0x03) << 4) | 
					   ( itemp[index+1] >> 4)];
				*optr++ = UUEncodeTable[((itemp[index+1] & 0x0f) << 2)];
				*optr++ = UUEncodeTable[0];
			} else if (count - index == 1) {
				*optr++ = UUEncodeTable[ itemp[index] >> 2];
				*optr++ = UUEncodeTable[(itemp[index] & 0x03) << 4];
				*optr++ = UUEncodeTable[0];
				*optr++ = UUEncodeTable[0];
			}
		}
	
		/* end of line */
		*optr++ = '\n';
		*optr++ = '\0';
	
		g_string_append (part, otemp);

		line++;
	}
	
	r_str = part->str;

	if (strlen (r_str) == 0) {
		g_string_free (part, TRUE);
		return NULL;
	} else {
		g_string_free (part, FALSE);
		return r_str;
	}
}
