/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#include <glib.h>
#include <gnome.h>

#include "file-grouplist.h"
#include "group.h"
#include "prefs.h"
#include "server.h"
#include "util.h"

void
file_grouplist_destroy (const Server * server)
{
	gchar * path;

	g_return_if_fail (server != NULL);

	path = g_strdup_printf ("/%s/%s.idx", data_dir, server->name);
	if (file_exists (path)) {
		g_message (_("Deleting file \"%s\""), path);
		remove (path);
	}
	g_free (path);

	path = g_strdup_printf ("/%s/%s.dat", data_dir, server->name);
	if (file_exists (path)) {
		g_message (_("Deleting file \"%s\""), path);
		remove (path);
	}
	g_free (path);

	path = g_strdup_printf ("/%s/%s", data_dir, server->name);
	if (file_exists (path)) {
		g_message (_("Deleting file \"%s\""), path);
		remove (path);
	}
	g_free (path);
}

gboolean
file_grouplist_exists (const Server* server)
{
	gchar * path = g_strdup_printf ("/%s/%s.idx", data_dir, server->name);
	gboolean retval = file_exists (path);
	g_free (path);
	return retval;
}

void
file_grouplist_load (Server * server, StatusItem * status)
{
	gchar * path;
	struct timeval start;
	struct timeval finish;
	double diff;
	GArray *dat, *idx;
	gboolean success = FALSE;

	gettimeofday (&start, NULL);

	/* open the index file */
	path = g_strdup_printf ("/%s/%s.idx", data_dir, server->name);
	idx = read_file (path);
	g_free (path);

	/* open the data file */
	path = g_strdup_printf ("/%s/%s.dat", data_dir, server->name);
	dat = read_file (path);
	g_free (path);

	if (dat!=NULL && idx!=NULL)
	{
		const gchar * dat_str = dat->data;
		const gchar * march_idx = idx->data;
		const glong version = get_next_token_int (march_idx, '\n', &march_idx);

		if (version == 1)
		{
			glong l;
			gint i, qty = get_next_token_int (march_idx, '\n', &march_idx);

			server->one_big_chunk = dat->data;
			pan_g_ptr_array_reserve (server->groups, qty);

			if (status != NULL)
				status_item_emit_init_steps (status, qty);

			for (i=0; i!=qty; ++i)
			{
				const gchar * newsrc = NULL;
				Group * g = group_new (server, dat_str+get_next_token_ulong(march_idx,'\n',&march_idx));
				if ((l = get_next_token_long (march_idx, '\n', &march_idx)) != -1)
					g->readable_name = dat_str+l;
				if ((l = get_next_token_long (march_idx, '\n', &march_idx)) != -1)
					g->description = dat_str+l;
				if ((l = get_next_token_long (march_idx, '\n', &march_idx)) != -1)
					g->download_dir = dat_str+l;
				if ((l = get_next_token_long (march_idx, '\n', &march_idx)) != -1)
					newsrc = dat_str+l;
				g->flags            = get_next_token_ulong (march_idx, '\n', &march_idx);
				g->state_filter     = get_next_token_ulong (march_idx, '\n', &march_idx);
				g->sort_style       = get_next_token_int   (march_idx, '\n', &march_idx);
				g->article_low      = get_next_token_ulong (march_idx, '\n', &march_idx);
				g->article_high     = get_next_token_ulong (march_idx, '\n', &march_idx);
				g->article_high_old = g->article_high;
				g->article_qty      = get_next_token_int   (march_idx, '\n', &march_idx);
				g->article_read_qty = get_next_token_int   (march_idx, '\n', &march_idx);
				newsrc_init (g->newsrc, newsrc, g->article_low, g->article_high);
				g_ptr_array_add (server->groups, g);

				if (status != NULL) {
					status_item_emit_next_step (status);
					if (!(server->groups->len % 512))
						status_item_emit_status (status,
							_("Loading %u of %d groups"), server->groups->len, qty);
				}
			}

			success = TRUE;
		}
		else
		{
			pan_error_dialog (_("Unsupported data version: %d"), version);
		}
	}
	
	gettimeofday (&finish, NULL);
	diff = finish.tv_sec - start.tv_sec;
	diff += (finish.tv_usec - start.tv_usec)/1000000.0;
	g_message (_("Got %d groups in %.1f seconds (%.0f groups/sec)"),
		server->groups->len,
		diff,
		server->groups->len/(fabs(diff)<0.001?0.001:diff));

	if (dat!=NULL) g_array_free (dat, !success);
	if (idx!=NULL) g_array_free (idx, TRUE);
}

void
file_grouplist_save (Server * server, StatusItem * status)
{
	guint i;
	FILE * idx_fp;
	FILE * dat_fp;
	gchar * pch;
	gchar * idx_path;
	gchar * dat_path;
	long pos = 0;
	struct timeval start;
	struct timeval finish;
	double diff;

	g_return_if_fail (server!=NULL);
	g_return_if_fail (server->groups!=NULL);
	g_return_if_fail (is_nonempty_string(data_dir));

	/* trivial case #1: no change, so no need to save */
	if (server->groups_dirty == FALSE)
		return;

	/* trivial case #2: nothing to save */
	if (server->groups->len == 0) {
		file_grouplist_destroy (server);
		return;
	}

	/**
	***  Save the Groups
	**/

	gettimeofday (&start, NULL);

	/* open index file */
	idx_path = g_strdup_printf ("/%s/%s.idx.tmp", data_dir, server->name);
	idx_fp = fopen (idx_path, "w+");
	if (idx_fp == NULL) {
		g_free (idx_path);
		return;
	}

	/* open data file */
	dat_path = g_strdup_printf ("/%s/%s.dat.tmp", data_dir, server->name);
	dat_fp = fopen (dat_path, "w+");
	if (dat_fp == NULL) {
		fclose (idx_fp);
		remove (idx_path);
		g_free (idx_path);
		g_free (dat_path);
		return;
	}

	/* Write DATBASE_VERSION */
	fprintf (idx_fp, "1\n%u\n", server->groups->len);

	/* Write the group information... */
	pos = 0;
	for (i=0; i!=server->groups->len; ++i)
	{
		Group * group = GROUP(g_ptr_array_index (server->groups, i));
		gchar * newsrc = newsrc_get_read_str(group->newsrc);
		long name_idx, read_idx, desc_idx, path_idx, news_idx;

		/* write the group name */
		name_idx = pos;
		pos += fwrite (group->name, 1, strlen(group->name)+1, dat_fp);

		/* write the readable name */
		if (((read_idx = (is_nonempty_string(group->readable_name) ? pos : -1))) != -1)
			pos += fwrite (group->readable_name, 1, strlen(group->readable_name)+1, dat_fp);

		/* write the description */
		if (((desc_idx = (is_nonempty_string(group->description) ? pos : -1))) != -1)
			pos += fwrite (group->description, 1, strlen(group->description)+1, dat_fp);

		/* write the download dir */
		if (((path_idx = (is_nonempty_string(group->download_dir)) ? pos : -1)) != -1)
			pos += fwrite (group->download_dir, 1, strlen(group->download_dir)+1, dat_fp);

		/* write the newsrc */
		if (((news_idx = (is_nonempty_string(newsrc) ? pos : -1))) != -1)
			pos += fwrite (newsrc, 1, strlen(newsrc)+1, dat_fp);

		/* write the non-string fields. */
		fprintf (idx_fp,
			"%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n"
			"%u\n" "%u\n" "%u\n" "%lu\n" "%lu\n" "%d\n" "%d\n",
			name_idx, read_idx, desc_idx, path_idx, news_idx,
			(group->flags & ~GROUP_NEW),
			group->state_filter,
			group->sort_style,
			group->article_low,
			group->article_high,
			group->article_qty,
			group->article_read_qty);

		/* let the user know what's going on */
		if (status != NULL) {
			status_item_emit_next_step (status);
			if (!(i % 512))
				status_item_emit_status (status,
					_("Saved %d of %u articles"),
					i, server->groups->len);
		}

		/* cleanup */
		g_free (newsrc);
	}

	/* the write went okay; update the idx file */
	fclose (idx_fp);
	pch = g_strdup_printf ("/%s/%s.idx", data_dir, server->name);
	remove (pch);
	rename (idx_path, pch);
	g_free (idx_path);
	g_free (pch);

	/* the write went okay; update the dat file */
	fclose (dat_fp);
	pch = g_strdup_printf ("/%s/%s.dat", data_dir, server->name);
	remove (pch);
	rename (dat_path, pch);
	g_free (dat_path);
	g_free (pch);

	/* cleanup */
	server->groups_dirty = FALSE;

	/* finish timing stats */
	gettimeofday (&finish, NULL);
	diff = finish.tv_sec - start.tv_sec;
	diff += (finish.tv_usec - start.tv_usec)/1000000.0;
	g_message (_("Saved %d groups in \"%s\" in %.1f seconds (%.0f groups/sec)"),
		server->groups->len,
		server->name,
		diff,
		server->groups->len/(fabs(diff)<0.001?0.001:diff));
}
