/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>

#include <glib.h>
#include <gnome.h>

#include "article.h"
#include "file-headers.h"
#include "gnksa.h"
#include "group.h"
#include "status-item.h"
#include "prefs.h"
#include "util.h"

/***
****
****
****
***/

gboolean
file_headers_exists (const Group * group)
{
	gchar * fname = g_strdup_printf ("/%s/%s/%s.idx",
		data_dir,
		group->server->name,
		group->name);
	gboolean retval = file_exists (fname);
	g_free (fname);
	return retval;
}

void
file_headers_load (Group * group, StatusItem * status)
{
	gboolean success = FALSE;
        gchar * path;
	struct timeval start;
	struct timeval finish;
	double diff;
	GArray *dat, *idx;

	gettimeofday (&start, NULL);

	/* open the index file */
	path = g_strdup_printf ("/%s/%s/%s.idx", data_dir, group->server->name, group->name);
	idx = read_file (path);
	g_free (path);

	/* open the data file */
	path = g_strdup_printf ("/%s/%s/%s.dat", data_dir, group->server->name, group->name);
	dat = read_file (path);
	g_free (path);

	if (dat!=NULL && idx!=NULL)
	{
		const gchar * dat_str = dat->data;
		const gchar * march = idx->data;
		const glong version = get_next_token_int (march, '\n', &march);

		if (version==1 || version==2 || version==3)
		{
			glong l;
			int i, qty = get_next_token_int (march, '\n', &march);
			GPtrArray * addme = g_ptr_array_new ();

			group->one_big_chunk = dat->data;
			pan_g_ptr_array_reserve (group->articles, qty);

			for (i=0; i!=qty; ++i)
			{
				Article * a = article_new (group);
				int extra_header_qty;
				int j;

				/* string fields */
				if ((l = get_next_token_long (march, '\n', &march)) != -1)
					a->message_id = dat_str+l;
				if (version<2) { /* version 2 split author into 2 fields */
					if ((l = get_next_token_long (march, '\n', &march)) != -1)
						article_set_author_from_header (a, dat_str+l);
				} else {
					if ((l = get_next_token_long (march, '\n', &march)) != -1)
						a->author_addr = dat_str+l;
					if ((l = get_next_token_long (march, '\n', &march)) != -1)
						a->author_real = dat_str+l;
				}
				if ((l = get_next_token_long (march, '\n', &march)) != -1)
					a->subject = dat_str+l;
				if (version<3) /* version 3 removed the date string */
					get_next_token_long (march, '\n', &march);
				if ((l = get_next_token_long (march, '\n', &march)) != -1)
					a->references = dat_str+l;

				/* numeric fields */
				a->part           = get_next_token_int (march, '\n', &march);
				a->parts          = get_next_token_int (march, '\n', &march);
				a->linecount      = get_next_token_int (march, '\n', &march);
				a->crosspost_qty  = get_next_token_int (march, '\n', &march);
				a->state          = get_next_token_int (march, '\n', &march);
				a->date           = get_next_token_ulong (march, '\n', &march);
				a->number         = get_next_token_ulong (march, '\n', &march);

				g_assert (a->number != 0);

				/* extra headers */
				extra_header_qty  = get_next_token_int (march, '\n', &march);
				for (j=0; j<extra_header_qty; ++j) {
					const gchar * key = dat_str + get_next_token_long(march,'\n',&march);
					const gchar * val = dat_str + get_next_token_long(march,'\n',&march);
					article_set_header_pair_chunk  (a, key, val);

				}

				/* let the user know what we're doing */
				if (status != NULL) {
					status_item_emit_next_step (status);
					if (!(addme->len % 256))
						status_item_emit_status (status,
							_("Loaded %d of %d articles"), i, qty);
				}

				/* add the article to the group */
				g_ptr_array_add (addme, a);
			}

			group_add_articles (group, addme, status);
			group->articles_dirty = FALSE;
			success = TRUE;
			g_ptr_array_free (addme, TRUE);
		}
		else
		{
			pan_error_dialog (_("Unsupported data version: %d"), version);
		}
	}

	/* expire the old articles, if any */
	if (success)
		group_expire_old_articles (group);

	/* timing stats */
	gettimeofday (&finish, NULL);
	diff = finish.tv_sec - start.tv_sec;
	diff += (finish.tv_usec - start.tv_usec)/1000000.0;
	g_message (_("Got %d articles in %.1f seconds (%.0f art/sec)"),
		group->articles->len,
		diff,
		group->articles->len/(fabs(diff)<0.001?0.001:diff));

	if (dat != NULL) g_array_free (dat, !success);
	if (idx != NULL) g_array_free (idx, TRUE);
}

/**
 * This is used to make sure we don't write the same string to the output
 * file more than once.  This way we can get some space savings from authors
 * and subjects that are repeated many times.
 */
static gint
get_string_offset (GHashTable * hash,
                   const gchar * str,
                   FILE * fp,
                   gint * pos)
{
	gpointer p;
	gint retval;

	g_assert (fp != NULL);
	g_assert (pos != NULL);

	if (!is_nonempty_string(str)) /* nothing to write */
	{
		retval = -1;
	}
	else if (hash == NULL) /* don't bother weeding duplicates */
	{
		retval = *pos;
		*pos += fwrite (str, 1, strlen(str)+1, fp);
	}
	else if ((p = g_hash_table_lookup(hash, str)) != NULL) /* a match! */
	{
		retval = GPOINTER_TO_INT(p);
	}
	else
	{
		retval = *pos;
		g_hash_table_insert (hash, (gpointer)str, GINT_TO_POINTER(retval));
		*pos += fwrite (str, 1, strlen(str)+1, fp);
	}

	return retval;
}

void
file_headers_save (Group * group, StatusItem * status)
{
	struct timeval start;
	struct timeval finish;
	double diff;
	guint i;
	FILE * idx_fp;
	FILE * dat_fp;
	gchar * pch;
	gchar * idx_path;
	gchar * dat_path;
	gint pos = 0;
	GHashTable * hash = NULL;

	g_return_if_fail (group!=NULL);
	g_return_if_fail (group->server!=NULL);
	g_return_if_fail (group->articles!=NULL);

	gettimeofday (&start, NULL);

	/* trivial case #1: no change, so no need to save */
	if (group->articles_dirty == FALSE)
		return;

	/* trivial case #2: nothing to save */
	if (group->articles->len == 0) {
		file_headers_destroy (group);
		return;
	}

	/**
	***  Save the Headers
	**/

	g_message (_("Saving \"%s\""), group_get_readable_name(group));

	/* open index file */
	idx_path = g_strdup_printf ("/%s/%s/%s.idx.tmp", data_dir, group->server->name, group->name);
	idx_fp = fopen (idx_path, "w+");
	if (idx_fp == NULL) {
		g_free (idx_path);
		return;
	}

	/* open data file */
	dat_path = g_strdup_printf ("/%s/%s/%s.dat.tmp", data_dir, group->server->name, group->name);
	dat_fp = fopen (dat_path, "w+");
	if (dat_fp == NULL) {
		fclose (idx_fp);
		remove (idx_path);
		g_free (idx_path);
		g_free (dat_path);
		return;
	}

	/* Write DATBASE_VERSION */
	fprintf (idx_fp, "3\n%u\n", group->articles->len);

	/* Write the article information... */
	pos = 0;
	hash = g_hash_table_new (g_str_hash, g_str_equal);
	for (i=0; i!=group->articles->len; ++i)
	{
		Article * a = ARTICLE(g_ptr_array_index (group->articles, i));
		long id_idx, author_real_idx, author_addr_idx, subj_idx, refs_idx;
		GPtrArray * extra_headers = article_get_all_headers (a);
		guint j;

		id_idx          = get_string_offset (NULL, a->message_id,  dat_fp, &pos);
		author_addr_idx = get_string_offset (hash, a->author_addr, dat_fp, &pos);
		author_real_idx = get_string_offset (hash, a->author_real, dat_fp, &pos);
		subj_idx        = get_string_offset (hash, a->subject,     dat_fp, &pos);
		refs_idx        = get_string_offset (NULL, a->references,  dat_fp, &pos);

		g_assert (a->number != 0);

		/* write the non-string fields. */
		fprintf (idx_fp,
			"%ld\n" "%ld\n" "%ld\n" "%ld\n" "%ld\n"
			"%d\n" "%d\n"
			"%d\n"
			"%d\n"
			"%d\n"
			"%lu\n"
			"%lu\n",
			id_idx, author_addr_idx, author_real_idx, subj_idx, refs_idx,
			a->part, a->parts,
			a->linecount,
			a->crosspost_qty,
			a->state,
			a->date,
			a->number);

		/* write the extra headers */
		g_assert (!(extra_headers->len % 2));
		fprintf (idx_fp, "%d\n", extra_headers->len/2); /* extra header qty */
		for (j=0; j!=extra_headers->len; j+=2)
		{
			const gchar * key = (const gchar*) g_ptr_array_index (extra_headers, j);
			const gchar * val = (const gchar*) g_ptr_array_index (extra_headers, j+1);

			/* write the key */
			fprintf (idx_fp, "%d\n", pos);
			pos += fwrite (key, 1, strlen(key)+1, dat_fp);

			/* write the val */
			fprintf (idx_fp, "%d\n", pos);
			pos += fwrite (val, 1, strlen(val)+1, dat_fp);
		}

		g_ptr_array_free (extra_headers, TRUE);
	}
	g_hash_table_destroy (hash);
	hash = NULL;

	/* the write went okay; move the idx file over */
	fclose (idx_fp);
	pch = g_strdup_printf ("/%s/%s/%s.idx", data_dir, group->server->name, group->name);
	remove (pch);
	rename (idx_path, pch);
	g_free (idx_path);
	g_free (pch);

	/* the write went okay; move the dat file over */
	fclose (dat_fp);
	pch = g_strdup_printf ("/%s/%s/%s.dat", data_dir, group->server->name, group->name);
	remove (pch);
	rename (dat_path, pch);
	g_free (dat_path);
	g_free (pch);

	/* cleanup */
	group->articles_dirty = FALSE;

	/* timing stats */
	gettimeofday (&finish, NULL);
	diff = finish.tv_sec - start.tv_sec;
	diff += (finish.tv_usec - start.tv_usec)/1000000.0;
	g_message (_("Saved %d articles in %.1f seconds (%.0f art/sec)"),
		group->articles->len,
		diff,
		group->articles->len/(fabs(diff)<0.001?0.001:diff));
}

void
file_headers_destroy (const Group * group)
{
	gchar * path;

	g_return_if_fail (group!=NULL);
	g_return_if_fail (group->server!=NULL);
	g_return_if_fail (is_nonempty_string(data_dir));

	path = g_strdup_printf ("/%s/%s/%s.idx", data_dir, group->server->name, group->name);
	g_message (_("Deleting file \"%s\""), path);
	remove (path);
	g_free (path);

	path = g_strdup_printf ("/%s/%s/%s.dat", data_dir, group->server->name, group->name);
	g_message (_("Deleting file \"%s\""), path);
	remove (path);
	g_free (path);
}
