/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <string.h>
#include <stdlib.h>

#include <glib.h>
#include <gnome.h>

#include "acache.h"
#include "article-thread.h"
#include "articlelist.h"
#include "debug.h"
#include "file-headers.h"
#include "globals.h"
#include "group.h"
#include "grouplist.h"
#include "log.h"
#include "newsrc.h"
#include "nntp.h"
#include "pan-callback.h"
#include "prefs.h"
#include "gui.h"
#include "queue.h"
#include "server.h"
#include "task-bodies.h"
#include "task-headers.h"
#include "task-grouplist.h"
#include "task-func-ptr.h"
#include "util.h"

PanCallback * group_articles_removed = NULL;

PanCallback * group_articles_added = NULL;

void
group_module_init (void)
{
	group_articles_removed = pan_callback_new ();
	group_articles_added = pan_callback_new ();
}

Group*
group_new (Server * server, const char * name)
{
	Group * g = g_new0 (Group, 1);
	debug1 (DEBUG_PAN_OBJECT, "group_new: %p", g);
        group_constructor (g, group_destructor, server, name);
	return g;
}

void
group_constructor (Group                * g,
                   PanObjectDestructor    dtor,
		   Server               * server,
		   const char           * name)
{
	g_assert (server != NULL);
	pan_warn_if_fail (is_nonempty_string(name));

	pan_object_constructor (PAN_OBJECT(g), dtor);

	debug1 (DEBUG_PAN_OBJECT, "group_constructor: %p", g);
	g->server = server;
	group_set_string_field (g, &g->name, name);
	g->readable_name = NULL;
	g->description = NULL;
	g->download_dir = NULL;
	g->flags = 0;
	g->state_filter = ~STATE_FILTER_IGNORED;
	g->article_low = (gulong)0;
	g->article_high_old = (gulong)0;
	g->article_high = (gulong)0;
	g->article_qty = (gint)0;
	g->article_read_qty = (gint)0;
	g->newsrc = newsrc_new (NULL, g->article_low, g->article_high);
	g->sort_style = 0;
	g->articles = g_ptr_array_new ();
	g->articles_refcount = 0;
	g->articles_dirty = FALSE;
	g->articles_threading_dirty = FALSE;
	g->chunk = NULL;
	g->one_big_chunk = NULL;
}

void
group_destructor (PanObject * o)
{
	Group * g = GROUP(o);

	g_return_if_fail (o!=NULL);

	debug1 (DEBUG_PAN_OBJECT, "group_destructor: %p", g);

	/* newsrc */
	pan_object_unref (PAN_OBJECT(g->newsrc));

	/* articles */
	pan_warn_if_fail (g->articles_refcount == 0);
	g_free (g->one_big_chunk);
	if (g->chunk != NULL)
		g_string_chunk_free (g->chunk);
	g_ptr_array_free (g->articles, TRUE);
	g->articles = NULL;

	pan_object_destructor (o);
}

/***
****
****  PUBLIC ACCESSORS / MUTATORS
****
***/

void
group_set_new (Group * group, gboolean is_new)
{
	g_return_if_fail (group != NULL);

	if (is_new)
		group_set_flags (group, group->flags|GROUP_NEW);
	else
		group_set_flags (group, group->flags&~GROUP_NEW);
}

gboolean 
group_is_new (const Group * group)
{
	g_return_val_if_fail (group!=NULL, FALSE);
	return (group->flags & GROUP_NEW) ? 1 : 0;
}

void
group_set_subscribed (Group * group, gboolean subscribed)
{
	g_return_if_fail (group != NULL);

	if (subscribed)
	{
		group_set_flags (group, group->flags|GROUP_SUBSCRIBED);
	}
	else
	{
		group_set_flags (group, group->flags&~GROUP_SUBSCRIBED);

		if (group->article_qty != 0)
			group_empty_dialog (group);
	}
}

gboolean 
group_is_subscribed (const Group * group)
{
	g_return_val_if_fail (group!=NULL, FALSE);

	return (group->flags & GROUP_SUBSCRIBED) ? 1 : 0;
}

void
group_set_dirty (Group * group)
{
	g_return_if_fail (group != NULL);
	group->server->groups_dirty = TRUE;
}

void
group_set_download_dir (Group * group, const gchar* download_dir)
{
	group_set_string_field (group, &group->download_dir, download_dir);
}

void
group_set_flags (Group * group, guint flags)
{
	group->flags = flags;
	group_set_dirty (group);
}

void
group_set_state_filter (Group * group, guint state_filter)
{
	group->state_filter = state_filter;
	group_set_dirty (group);
}

void
group_set_sort_style (Group * group, int sort_style)
{
	group->sort_style = sort_style;
	group_set_dirty (group);
}

void
group_set_article_range (Group * group,
                         gulong article_low,
			 gulong article_high)
{
	group->article_low = article_low;
	group->article_high = article_high;
	group_set_dirty (group);
}

void
group_mark_new_article_number (Group         * group,
                               gulong          everything_above_this_is_new)
{
	group->article_high_old = everything_above_this_is_new;
}

void
group_set_article_qty (Group * group, gint article_qty)
{
	g_return_if_fail (group != NULL);

	if (article_qty != group->article_qty)
	{
		group->article_qty = article_qty;
		group_set_dirty (group);
		grouplist_update_row (group);
	}
}
void
group_inc_article_qty (Group * group, gint inc)
{
	g_return_if_fail (group != NULL);

	group_set_article_qty (group, group->article_qty + inc);
}

void
group_set_article_read_qty (Group * group, gint article_read_qty)
{
	article_read_qty = MAX (article_read_qty, 0);
	group->article_read_qty = article_read_qty;
	group_set_dirty (group);
	grouplist_update_row (group);
}
void
group_inc_article_read_qty (Group * group, gint inc)
{
	group_set_article_read_qty (group, group->article_read_qty + inc);
}

void
group_mark_all_read (Group * g, gboolean read)
{
	g_return_if_fail (g != NULL);

	newsrc_set_group_range (g->newsrc, g->article_low, g->article_high);
	newsrc_mark_all (g->newsrc, read);
	group_set_article_read_qty (g, (read ? g->article_qty : 0));
}

gboolean
group_is_folder (const Group * group)
{
	g_return_val_if_fail (group!=NULL, FALSE);
	return group->flags & GROUP_FOLDERS ? 1 : 0;
}

void
group_set_is_folder (Group * group, gboolean folder)
{
	if (!folder != !group_is_folder(group))
	{
		if (folder)
			group->flags |= GROUP_FOLDERS;
		else
			group->flags &= ~GROUP_FOLDERS;
		group_set_dirty (group);
	}
}

/***
****
****  DELETE GROUP
****
***/

/* empty the group, delete the group database file */
static void
real_group_empty (Group * group, gboolean clear_counts)
{
	g_return_if_fail (group != NULL);
	g_return_if_fail (group->articles != NULL);

	/* unload group (if it's loaded) */
	if (articlelist_get_group() == group)
		articlelist_set_group (NULL);

	/* clear out the article headers */
	file_headers_destroy (group);

	/* wipe out all the articles... */
	pan_g_ptr_array_foreach (group->articles, (GFunc)pan_object_unref, NULL);
	g_ptr_array_set_size (group->articles, 0);
	group_set_article_qty (group, 0);
	group_set_article_read_qty (group, 0);
	if (clear_counts)
		group_set_article_range (group, 0, 0);

	/* flush the flags */
	if (group->flags & GROUP_SUBSCRIBED)
		group_set_flags (group, GROUP_SUBSCRIBED);
	else
		group_set_flags (group, 0);
}

/* callback for button clicks in group_empty_dialog */
static void
group_empty_cb (GnomeDialog *dialog,
		int button_number,
		gpointer group)
{
	switch (button_number)
	{
		case 0: /* 'yes' */
		{
			GtkWidget * w = gtk_object_get_data (GTK_OBJECT(dialog), "reset");
			gboolean clear_counts = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(w));
			g_message ("clear_counts: %d", clear_counts);
			real_group_empty (group, clear_counts);
			break;
		}
		case 1: /* 'no' */
			break;

		default: /* cancel */
			break;
	};

	gtk_widget_destroy (GTK_WIDGET(dialog));
}

void
group_empty_dialog (Group *group)
{
	GString * str = g_string_new (NULL);
	GtkWidget * dialog = NULL;
	GtkWidget * w = NULL;

	/* sanity check */
	g_return_if_fail (group!=NULL);

	pan_lock();

	/* create dialog */
	g_string_sprintf (str,
	                  _("Clean group \"%s\""),
	                  group_get_readable_name(group));
	dialog = gnome_dialog_new (str->str,
				   GNOME_STOCK_BUTTON_YES,
				   GNOME_STOCK_BUTTON_NO, NULL);

	/* body */
	g_string_sprintf (str,
	                  _("Remove all messages from group: \n%s ?"),
	                  group_get_readable_name(group));
	w = gtk_label_new (str->str);
	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), w);

	/* "clear count?" check button */
	w = gtk_check_button_new_with_label (_("Reset new message counter"));
	gtk_container_add (GTK_CONTAINER(GNOME_DIALOG(dialog)->vbox), w);
	gtk_object_set_data (GTK_OBJECT(dialog), "reset", w);

	/* listen for clicks */
	gtk_signal_connect (GTK_OBJECT(dialog), "clicked",
			    GTK_SIGNAL_FUNC (group_empty_cb), group);

	pan_unlock();

	gui_popup_draw (dialog, Pan.window);

	g_string_free (str, TRUE);
}


/***
****
****  CATCHUP
****
***/

static int
group_catchup_task_run (TaskFuncPtr * tfp, gpointer user_data)
{
	gint result;
	gulong qty=0, low=0, high=0;
	Group * group = GROUP (user_data);

	/* get the new group info */
	result = nntp_get_group_info (STATUS_ITEM(tfp),
	                              TASK(tfp)->sock,
	                              group->name,
	                              &qty, &low, &high,
	                              &TASK(tfp)->hint_abort);
	if (result != TASK_SUCCESS)
		return result;

	/* update our group with the info */
	group_set_article_qty (group, qty);
	group_set_article_range (group, low, high);
	newsrc_mark_all (group->newsrc, TRUE);
	group_set_article_read_qty (group, qty);
	return TASK_SUCCESS;
}

static char*
group_catchup_task_describe (const StatusItem * status)
{
	const TaskFuncPtr * ptr = TASK_FUNC_PTR (status);
	Group * group = GROUP (ptr->user_data);
	return g_strdup_printf (_("Catching up group `%s'"),
	                        group_get_readable_name(group));
}

void
group_catchup (Group * group)
{
	if (articlelist_get_group() == group)
	{
		articlelist_all_mark_read ();
	}
	else
	{
		queue_add (TASK(task_func_ptr_new (group->server,
						   group_catchup_task_describe,
						   group_catchup_task_run,
						   group,
						   NULL)));
	}
}

/***
****
****  UPDATE COUNT INFO
****
***/

static int
group_update_count_info_task_run (TaskFuncPtr * task, gpointer user_data)
{
	GPtrArray * groups = (GPtrArray*) user_data;
	StatusItem * status = STATUS_ITEM(task);
	gint result;
	gint retval = TASK_SUCCESS;
	guint i;

	/* for each group, get the numbers... */
	status_item_emit_init_steps (status, groups->len);
	for (i=0; i!=groups->len; ++i)
	{
		gulong qty=0, low=0, high=0;
		Group * g = GROUP(g_ptr_array_index(groups,i));

		/* let the status item know what we're doing */
		status_item_emit_next_step (status);
		if (!(i % 10))
			status_item_emit_status (status,
				_("Updated counts for %u groups out of %u"),
				i, groups->len);

		/* get the numbers */
		result = nntp_get_group_info (STATUS_ITEM(task),
		                              TASK(task)->sock,
		                              g->name,
		                              &qty, &low, &high,
		                              &TASK(task)->hint_abort);
		if (result != TASK_SUCCESS) {
			retval = result;
			break;
		}

		/* update the group object */
		group_set_article_range (g, low, high);
		group_set_article_qty (g, qty);
		group_set_article_read_qty (
			g, g->article_read_qty-(low-g->article_low));
	}

	return retval;
}

static char*
group_update_count_info_task_describe (const StatusItem * status)
{
	return g_strdup_printf (
		_("Updating article counts for %u groups from server `%s'"),
		((GPtrArray*)(TASK_FUNC_PTR(status)->user_data))->len,
		TASK(status)->server->name);
}

static void
group_update_count_dtor (gpointer data)
{
	/* free the gptrarray passed into group_update_count_info */
	TaskFuncPtr * task = (TaskFuncPtr*) data;
	GPtrArray * a = (GPtrArray*) task->user_data;
	g_ptr_array_free ((GPtrArray*)a, TRUE);
}

void
group_update_count_info (GPtrArray * groups)
{
	Server * server;

	g_return_if_fail (groups != NULL);
	g_return_if_fail (groups->len != 0);

	/* the task owns & will free 'groups' now... */
	server = GROUP(g_ptr_array_index(groups,0))->server;
	queue_add (TASK(task_func_ptr_new(server,
	                                  group_update_count_info_task_describe,
	                                  group_update_count_info_task_run,
	                                  groups,
					  group_update_count_dtor)));
}


/**
***
**/

/**
***  ARTICLE LIST
**/

static int
compare_particle_pparticle_msgid (const void * a, const void * b)
{
	return strcmp (((const Article*)a)->message_id,
	              (*(const Article**)b)->message_id);
}

static void
group_add_articles_ghfunc (gpointer key, gpointer val, gpointer data)
{
	guint i;
	Group * group = GROUP(key);
	GPtrArray * articles = (GPtrArray*)val;

	/* this is so that unreffing from the old group doesn't kill
	   the article -- this ref, plus group_remove_article's unref,
	   has a net change of zero */
	for (i=0; i!=articles->len; ++i)
		pan_object_ref (PAN_OBJECT(g_ptr_array_index(articles,i)));

	/* remove the articles from the other group */
	group_remove_articles (group, articles);

	/* clean up */
	g_ptr_array_free (articles, TRUE);
}

void
group_add_article (Group * group, Article * article)
{
	GPtrArray * a = g_ptr_array_new ();
	g_ptr_array_add (a, article);
	group_add_articles (group, a, NULL);
	g_ptr_array_free (a, TRUE);
}

void
group_add_articles (Group * group, GPtrArray * articles, StatusItem * status)
{
	gint read_qty;
	guint i;
	gulong low;
	gulong high;
	gboolean reffed = FALSE;
	GHashTable * remove_from_old = NULL;

	/* sanity checks */
	g_return_if_fail (group!=NULL);
	g_return_if_fail (group->articles_refcount>=0);
	g_return_if_fail (articles!=NULL);
	if (!articles->len) return;
	g_return_if_fail (articles->len>0);

	g_message ("group_add_articles with %u articles", articles->len);


	/* warn the caller if they're doing something inefficient by adding
	   articles to a group whose refcount is zero.  The old articles in
	   this group have to be loaded too; otherwise they'll be overwritten
	   when we save this group to disk without them. */
	if (group->articles_refcount==0) {
		g_message ("group_add_articles with !group->articles_refcount");
		reffed = TRUE;
		group_ref_articles (group, NULL);
	}

	/**
	***  Remove the articles from their old groups;
	***  do this in batches for efficiency.  Sheesh.
	**/

	for (i=0; i!=articles->len; ++i)
	{
		Article * article = ARTICLE(g_ptr_array_index(articles,i));
		Group * oldgroup = article->group;
		if (oldgroup!=NULL && oldgroup!=group)
		{
			gboolean array_existed;
			GPtrArray * to_remove;

			/* ensure the hash table exists */
			if (remove_from_old == NULL)
				remove_from_old = g_hash_table_new (
					g_direct_hash, g_direct_equal);

			/* see if any articles are queued for removing
			   from that group */
			to_remove = g_hash_table_lookup (remove_from_old,
			                                 oldgroup);
			array_existed = to_remove != NULL;
			if (!array_existed)
				to_remove = g_ptr_array_new ();
			g_ptr_array_add (to_remove, article);
			if (!array_existed)
				g_hash_table_insert (remove_from_old,
				                     oldgroup,
				                     to_remove);
		}
	}
	if (remove_from_old != NULL) {
		g_hash_table_foreach (remove_from_old,
		                      group_add_articles_ghfunc,
		                      NULL);
		g_hash_table_destroy (remove_from_old);
		remove_from_old = NULL;
	}

	/**
	***  Now that the articles don't belong to another group,
	***  add them to this one.
	**/

	low = group->article_low;
	high = group->article_high;

	for (i=0; i!=articles->len;)
	{
		guint index;
		gboolean exact_match;
		Article * article;

		/* update the group's articles field */
		article = ARTICLE(g_ptr_array_index(articles,i));
		exact_match = FALSE;
		index = lower_bound (article,
		                     group->articles->pdata,
		                     group->articles->len,
		                     sizeof(gpointer),
		                     compare_particle_pparticle_msgid,
		                     &exact_match);
		if (exact_match)
		{
			g_ptr_array_remove_index (articles, i);

			/* we've already got this article...
			 * if it's a second copy of it, throw it away. */
			if (article != g_ptr_array_index(group->articles,index))
				pan_object_unref (PAN_OBJECT(article));
		}
		else
		{
			/* update the article fields */
			article->group = group;

			/* update the group fields */
			if (article->number < low)
				low = article->number;
			if (article->number > high)
				high = article->number;

			/* add the article */
			pan_g_ptr_array_insert (group->articles,
			                        article,
			                        index);

			++i;
		}
	}

g_message ("group_add_articles actually added %u articles", articles->len);

	/* update the article range */
	if (high > group->article_high || low<group->article_low)
		group_set_article_range (group, low, high);

	/* update the article fields */
	group->articles_dirty = TRUE;
	group_set_article_qty (group, group->articles->len);
	read_qty = 0;
	for (i=0; i!=group->articles->len; ++i)
		if (article_is_read (ARTICLE(g_ptr_array_index(group->articles,i))))
			++read_qty;
	group_set_article_read_qty (group, read_qty);

	/* let everyone know that articles are going to be added */
	if (articles->len != 0)
	{
		group->articles_threading_dirty = TRUE;
		pan_callback_call (group_articles_added, group, articles);
	}

	/* cleanup */
	if (reffed)
		group_unref_articles (group, NULL);

	/* folders are for permanent storage, so re-save *NOW*.
	   This is especially important for pan.sendlater. */
	if (group_is_folder(group))
		file_headers_save (group, status);
}

void
group_thread_if_needed (Group * group, StatusItem * status)
{
	g_return_if_fail (group != NULL);

	if (group->articles_threading_dirty && group->articles->len)
	{
		thread_articles (group->articles, status);
		check_multipart_articles (group->articles);
		group->articles_threading_dirty = FALSE;
	}
}

void
group_remove_article (Group * group, Article * article)
{
	GPtrArray * a = g_ptr_array_new ();
	g_ptr_array_add (a, article);
	group_remove_articles (group, a);
	g_ptr_array_free (a, TRUE);
}

void
group_remove_articles (Group * group, GPtrArray * articles)
{
	guint i;
	int removed_read = 0;
	int removed_total = 0;
	gboolean exact_match = FALSE;

	/* sanity checks */
	g_return_if_fail (group != NULL);
	g_return_if_fail (articles != NULL);
	if (!articles->len) return;
	for (i=0; i!=articles->len; ++i)
		g_assert (ARTICLE(g_ptr_array_index(articles,i))->group == group);

g_message ("group_remove_articles: removing %d articles from %s", (int)articles->len, group->name);

	/* let everyone know that articles are going to be removed */
	pan_callback_call (group_articles_removed, group, articles);

	/* is this the last gasp for these articles? */
	for (i=0; i!=articles->len; ++i) {
		Article * article = ARTICLE(g_ptr_array_index(articles,i));
		if (PAN_OBJECT(article)->ref_count == 1) { /* about to die */
			article_isolate (article);
			acache_delete (article->message_id);
		}
	}

	/* remove these articles */
	for (i=0; i!=articles->len; ++i)
	{
		int index;
		Article * article = ARTICLE(g_ptr_array_index(articles,i));

		pan_warn_if_fail (article != NULL);
		pan_warn_if_fail (article->group == group);

		/* find the article matching the message id */
		index = lower_bound (article,
				     group->articles->pdata,
				     group->articles->len,
				     sizeof(gpointer),
				     compare_particle_pparticle_msgid,
				     &exact_match);

		/* if we found one, remove it & update our stats */
		if (exact_match)
		{
			Article * a;

			++removed_total;
			if (article_is_read(article))
				++removed_read;

			/* update article fields */
			a = g_ptr_array_remove_index (group->articles, index);
			a->group = NULL;
			pan_object_unref (PAN_OBJECT(a));
		}
	}

	/* if articles were removed, update our stats */
	if (articles->len != 0)
	{
		group->articles_dirty = TRUE;
		group->article_qty -= removed_total;
		group->article_read_qty -= removed_read;
		grouplist_update_row (group);
		group->articles_threading_dirty = TRUE;
		if (group_is_folder(group))
			file_headers_save (group, NULL);
	}
}

void
group_expire_old_articles (Group * g)
{
	guint i;
	GPtrArray * removeme;

	g_return_if_fail (g != NULL);
	g_return_if_fail (g->articles != NULL);

	removeme = g_ptr_array_new ();

	/* find old articles to remove */
	for (i=0; i!=g->articles->len; ++i) {
		Article * a = ARTICLE(g_ptr_array_index(g->articles,i));
		if ((a->number<g->article_low) || (a->number>g->article_high))
			g_ptr_array_add (removeme, a);
	}

	/* remove the old articles */
	if (removeme->len != 0)
	{
		g_message ("expiring %u articles", removeme->len);
		group_remove_articles (g, removeme);
	}

	/* cleanup */
	g_ptr_array_free (removeme, TRUE);
}


/***
****
***/

void
group_ref_articles (Group * group, StatusItem * status)
{
	g_assert (group != NULL);
	g_assert (group->articles_refcount >= 0);

	g_message ("reffing; previous count was %d [%s]",
	           group->articles_refcount,
	           group->name);

	if (++group->articles_refcount == 1)
		file_headers_load (group, status);

	g_message ("group article refcount is %d [%s]", 
	           group->articles_refcount,
	           group->name);

	group_set_article_qty (group, group->articles->len);
}

void
group_unref_articles (Group * group, StatusItem * status)
{
	g_assert (group != NULL);
	g_assert (group->articles_refcount > 0);

	g_message ("unreffing; previous count was %d [%s]",
	           group->articles_refcount,
	           group->name);

	if (!--group->articles_refcount)
	{
		file_headers_save (group, status);

		/* free the articles */
		pan_g_ptr_array_foreach (group->articles,
		                         (GFunc)pan_object_unref,
		                         NULL);
		g_ptr_array_set_size (group->articles, 0);

		/* free the memory used by those articles */
		g_free (group->one_big_chunk);
		group->one_big_chunk = NULL;
		g_message ("clearing the string chunk for `%s'", group->name);
		if (group->chunk != NULL) {
			g_string_chunk_free (group->chunk);
			group->chunk = NULL;
		}
	}
}



/**
***
**/

Group*
folder_get_by_name (const gchar * foldername)
{
	return group_get_by_name (server_get_by_name(INTERNAL_SERVER_NAME),
	                          foldername);
}


const gchar*
group_get_readable_name (const Group * group)
{
	g_return_val_if_fail (group!=NULL, NULL);

	return is_nonempty_string (group->readable_name)
		? group->readable_name
		: group->name;
}

/**
***
**/

const gchar* group_chunk_string (Group * group,
                                 const gchar * chunkme)
{
	if (group->chunk == NULL) {
		g_message ("creating string chunk for group `%s'", group->name);
		group->chunk = g_string_chunk_new (4096);
	}

	return g_string_chunk_insert_const (group->chunk, chunkme);
}

void
group_set_string_field (Group         * group,
	                const gchar  ** field_ptr,
			const gchar   * chunkme_replacement)
{
	g_return_if_fail (group!=NULL);
	g_return_if_fail (group->server!=NULL);
	g_return_if_fail (field_ptr!=NULL);

	*field_ptr = server_chunk_string (group->server, chunkme_replacement);
}


/***
****
****  NEW FOLDER DIALOG
****
***/

static void
grouplist_selected_new_cb (gchar* string, gpointer data)
{
	Group * group = NULL;
	Server * server = server_get_by_name (INTERNAL_SERVER_NAME);
	gchar * tmp;

	/* user hit cancel */
	if (string == NULL)
		return;

	/* make sure the user entered something */
	tmp = g_strdup (string);
	g_strstrip (tmp);
	if (!is_nonempty_string (tmp)) {
		pan_error_dialog (_("Invalid folder name: \"%s\""), string);
		g_free (tmp);
		return;
	}

	/* make sure there's not already a folder with that name */
	g_strdown (tmp);
	group = folder_get_by_name (tmp);
	if (group != NULL) {
		pan_error_dialog (_("Folder \"%s\" already exists."), string);
		g_free (tmp);
		return;
	}

	/* make sure it's not in the reserved namespace */
	if (!strncmp(tmp, "pan.", 4)) {
		pan_error_dialog (_("Please don't begin folders with \"pan.\"; it confuses me terribly."));
		g_free (tmp);
		return;
	}

	/* create new folder */
	group = group_new (server, tmp);
	group_set_is_folder (group, TRUE);
	group_set_dirty (group);
	server_add_group (server, group);
	grouplist_update_ui (server);

	/* cleanup */
	g_free (tmp);
}

void
group_new_folder_dialog (void)
{
	gchar * prompt = g_strdup (_("New Folder Name:"));
	gnome_request_dialog (FALSE, prompt, NULL, 128,
			      grouplist_selected_new_cb,
			      NULL, GTK_WINDOW(Pan.window));
	g_free (prompt);
}

void
group_download_all_articles (Group * group)
{
	g_return_if_fail (group != NULL);
	g_return_if_fail (group->articles != NULL);

	group_ref_articles (group, NULL);
	queue_add (TASK(task_bodies_new (group, group->articles)));
	group_unref_articles (group, NULL);
}

