/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#ifndef __GROUPS_H__
#define __GROUPS_H__

#include <glib.h>

#include "newsrc.h"
#include "pan-object.h"
#include "server.h"
#include "status-item.h"

enum
{
	GROUP_ALL				= (1<<0),
	GROUP_SUBSCRIBED			= (1<<1), 
	GROUP_NEW				= (1<<2),
	GROUP_FOLDERS				= (1<<3)
};

enum
{
        STATE_FILTER_UNREAD                     = (1<<0),
        STATE_FILTER_READ                       = (1<<1),
        STATE_FILTER_WATCHED                    = (1<<2),
        STATE_FILTER_IGNORED                    = (1<<3),
        STATE_FILTER_NORMAL_RANK                = (1<<4),
        STATE_FILTER_COMPLETE_BINARIES          = (1<<5),
        STATE_FILTER_INCOMPLETE_BINARIES        = (1<<6),
        STATE_FILTER_NONBINARIES                = (1<<7),
        STATE_FILTER_SAVED                      = (1<<8),
        STATE_FILTER_QUEUED                     = (1<<9),
        STATE_FILTER_IDLE                       = (1<<10),
	STATE_FILTER_NEW			= (1<<11)
};

#define GROUP(x) ((Group*)(x))

/**
 * For the time being, groups are 'owned' by their server.
 * When we go group-centric, this will change.
 *
 * Having Group a handle to its server lets us pass around
 * groups w/o a server argument for context, which should make
 * refitting for groupcentric easier...
 */
typedef struct _Group
{
	PanObject      parent;                           /* parent class */

        gint           articles_dirty           : 1;     /* do the articles need saving? */
	gint           articles_threading_dirty : 1;
	gint           articles_refcount;
	gint           sort_style;                       /* (read only) hint for sorting in UI */
	gint           article_qty;                      /* (read only) # of articles in group */
	gint           article_read_qty;                 /* (read only) # of read articles */
	guint          flags;                            /* (read only) new, subscribed, etc */
	guint          state_filter;                     /* (read only) for filtering articles */
	gulong         article_low;                      /* (read only) low article number */
	gulong         article_high;                     /* (read only) high article number */
	gulong         article_high_old;                 /* (read only) previous high article */
	const gchar  * download_dir;                     /* where to save files, or NULL */
	Newsrc       * newsrc;                           /* tracks read/unread */

	Server       * server;
	const gchar  * name;                             /* alt.what.is.my.name */
	const gchar  * readable_name;                    /* maybe null, user alias for group */
	const gchar  * description;                      /* maybe null */

	GPtrArray    * articles;                         /* articles */
	GStringChunk * chunk;                            /* string chunk used by articles */
	gchar        * one_big_chunk;                    /* used by loading local headers */
}
Group;


/**
***  PUBLIC
**/

void      group_module_init              (void);

Group*    group_new                      (Server        * server,
                                          const char    * name);

void      group_catchup                  (Group         * group);

void      group_update_count_info        (GPtrArray     * groups);

void      group_inc_article_qty          (Group         * group,
                                          gint            article_qty);

void      group_inc_article_read_qty     (Group         * group,
                                          gint            article_read_qty);

void      group_set_article_qty          (Group         * group,
                                          gint            article_qty);

void      group_set_article_read_qty     (Group         * group,
                                          gint            article_read_qty);

void      group_set_article_range        (Group         * group,
                                          gulong          article_low,
                                          gulong          article_high);

void      group_mark_new_article_number  (Group         * group,
                                          gulong          everything_above_this_is_new);

void      group_mark_all_read            (Group         * group,
                                          gboolean        read);

void      group_set_string_field         (Group         * group,
                                          const gchar  ** field_ptr,
                                          const gchar   * replacement);

void      group_set_sort_style           (Group         * group,
                                          gint            sort_style);

void      group_set_state_filter         (Group         * group,
                                          guint           state_filter);

void      group_set_flags                (Group         * group,
                                          guint           flags);

void      group_set_download_dir         (Group         * group,
                                          const char    * download_dir);

gboolean  group_is_folder                (const Group   * group);

void      group_set_is_folder            (Group         * group,
                                          gboolean        is_folder);

const gchar* group_get_readable_name     (const Group   * group);


Group*    group_get_by_name              (Server        * server,
                                          const char    * groupname);

Group*    folder_get_by_name             (const char    * foldername);

void      group_empty_dialog             (Group         * group);

void      group_download_all_articles    (Group         * group);

void      group_set_subscribed           (Group         * group,
                                          gboolean        subscribed);

gboolean  group_is_subscribed            (const Group   * group);

void      group_set_new                  (Group         * group,
                                          gboolean        is_new);

gboolean  group_is_new                   (const Group   * group);

void      group_set_dirty                (Group         * group);

/**
 * ARTICLES
 *
 * the add and remove article functions all update the group's
 * statistics and the way it appears in the Pan user interface.
 * If you're adding or removing more than one groups, it's much
 * better to use the plural functions once instead of calling
 * the singular versions many times.
 */

struct _Article;

void      group_add_article              (Group            * group,
                                          struct _Article  * article);

void      group_add_articles             (Group            * group,
                                          GPtrArray        * articles,
                                          StatusItem       * status_or_null);

void      group_remove_article           (Group            * group,
                                          struct _Article  * article);

void      group_remove_articles          (Group            * group,
                                          GPtrArray        * articles);

void      group_ref_articles             (Group            * group,
                                          StatusItem       * item);

void      group_unref_articles           (Group            * group,
                                          StatusItem       * item);


void      group_new_folder_dialog        (void);

void      group_expire_old_articles      (Group            * group);


void      group_thread_if_needed         (Group            * group,
                                          StatusItem       * status);

const gchar* group_chunk_string (Group * group, const gchar * chunkme);

/**
***  PROTECTED
**/

void group_constructor    (Group                        * group,
                           PanObjectDestructor            dtor,
                           Server                       * server,
			   const gchar                  * name);

void group_destructor     (PanObject                    * group);


extern PanCallback * group_articles_removed;

extern PanCallback * group_articles_added;

extern const gchar * PAN_SENDLATER;
extern const gchar * PAN_SENT;

#endif /* __GROUPS_H__ */
