/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <gnome.h>

#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "message-check.h"
#include "gnksa.h"
#include "util.h"

/***
****  PRIVATE UTILITIES
***/


/* Find and return a list of NNTP groups to send to */
static GPtrArray*
get_nntp_rcpts (const gchar *to)
{
	const gchar * march;
	GPtrArray * a = g_ptr_array_new ();
	GString * rcpt = g_string_new (NULL);

	/* march through the tokens & save the nonempty ones */
	march = to;
	while (get_next_token_g_str (march, ',', &march, rcpt))
		g_ptr_array_add (a, g_strdup(rcpt->str));

	/* cleanup */
	g_string_free (rcpt, TRUE);

	return a;
}


/***
****  OUTGOING MESSAGE CHECKS
***/

/**
 * Check to see if the signature (if found) is within the McQuary limit of
 * four lines and 80 columns per line.
 */
static void
check_signature (GPtrArray * setme_errors,
                 GoodnessLevel * setme_goodness,
                 const char * body,
                 GString * line)
{
	int too_wide_qty = 0;
	int sig_line_qty = -1;

	g_return_if_fail (setme_errors!=NULL);
	g_return_if_fail (setme_goodness!=NULL);
	g_return_if_fail (body!=NULL);
	g_return_if_fail (line!=NULL);

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (strcmp(line->str,"-- ") && sig_line_qty<0)
			continue;

		++sig_line_qty;

		if (line->len>80)
			++too_wide_qty;
	}

	if (sig_line_qty == 0)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Signature prefix with no signature.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}

	else if (sig_line_qty > 4)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: signature is more than 4 lines long")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}

	if (too_wide_qty != 0)
	{
		g_ptr_array_add (setme_errors, g_strdup_printf(
			_("WARNING: %d lines too wide in signature"),
			too_wide_qty));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}


/**
 * Simple check to see if the body is too wide.  Any text after the
 * signature prefix is ignored in this test.
 */
static void
check_wide_body (GPtrArray * setme_errors,
                 GoodnessLevel * setme_goodness,
                 const gchar * body,
		 GString * line)
{
	int too_wide_qty = 0;

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (!strcmp (line->str, "-- "))
			break;
		if (line->len>80)
			++too_wide_qty;
	}

	if (too_wide_qty != 0)
	{
		g_ptr_array_add (setme_errors, g_strdup_printf(
			_("WARNING: %d lines >80 characters in the message."),
			too_wide_qty));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}
/**
 * Check to see if the article appears to be empty.  Any text after the
 * signature prefix is ignored in this test.
 */
static void
check_empty (GPtrArray * setme_errors,
             GoodnessLevel * setme_goodness,
             const gchar* body,
             GString * line)
{
	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line))
	{
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len != 0)
			return;
	}

	g_ptr_array_add (setme_errors, g_strdup(
		_("ERROR: Message is empty.")));
	*setme_goodness = MAX (*setme_goodness, REFUSE);
}

/**
 * Check to see how much original content is in this message, opposed
 * to quoted content.  Any text after the signature prefix is ignored
 * in this test.
 *
 * (1) count all the lines beginning with the quoted prefix.
 * (2) count all the nonempty nonsignature lines.  These are the orignal lines.
 * (3) if the ratio of original/quoted is 20% or less, warn.
 * (4) if the ratio of original/quoted is 0%, warn louder.
 */
static void
check_mostly_quoted (GPtrArray * setme_errors,
                     GoodnessLevel * setme_goodness,
                     const gchar * body,
                     GString * line)
{
	int unquoted = 0;
	int total = 0;

	g_string_truncate (line, 0);

	while (get_next_token_g_str (body, '\n', &body, line)) {
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len==0)
			continue;
		++total;
		if (*line->str != '>')
			++unquoted;
	}

	if (total!=0 && ((unquoted/(double)total)<0.2))
	{
		char * pch = g_strdup (unquoted==0
			?  _("WARNING: The message is entirely quoted text!")
			:  _("WARNING: The message is mostly quoted text."));
		g_ptr_array_add (setme_errors, pch);
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

/**
 * Check to see if the article appears to only have quoted text.  If this
 * appears to be the case, we will refuse to post the message.
 *
 * (1) Get mutable working copies of the article body and the attribution
 *     string.
 *
 * (2) Replace carriage returns in both the calculated attribution string
 *     and a temporary copy of the message body, so that we don't have to
 *     worry whether or not the attribution line's been wrapped.
 * 
 * (3) Search for an occurance of the attribution string in the body.  If
 *     it's found, remove it from the temporary copy of the body so that
 *     it won't affect our line counts.
 *
 * (4) Of the remaining body, look for any nonempty lines before the signature
 *     file that don't begin with the quote prefix.  If such a line is found,
 *     then the message is considered to not be all quoted text.
 *
 */
static void
check_all_quoted (GPtrArray      * setme_errors,
                  GoodnessLevel  * setme_goodness,
	          const gchar    * body,
                  GString        * line,
		  const Article  * article)
{
	char * body2;
	const char * pch;
	gchar * attribution;

	/* (1) */
	attribution = g_strdup (article_get_header (article, PAN_ATTRIBUTION));
	body2 = g_strdup (body);

	/* (2) */
	if (is_nonempty_string (attribution))
		g_strdelimit (attribution, "\n", ' ');
	g_strdelimit (body2, "\n", ' ');

	/* (3) */
	if (is_nonempty_string (attribution)) {
		int attribution_len = strlen (attribution);
		int attribution_start_index;
		pch = strstr (body2, attribution);
		attribution_start_index = pch==NULL ? -1 : pch-body2;
		g_free (body2);
		body2 = g_strdup (body);
		if (attribution_start_index != -1) {
			char * src = body2 + attribution_start_index + attribution_len;
			char * tgt = body2 + attribution_start_index;
			while (*src)
				*tgt++ = *src++;
			*tgt = '\0';
		}
	}
	g_free (attribution);

	/* (4) */
	pch = body2;
	while (get_next_token_g_str (pch, '\n', &pch, line)) {
		if (!strcmp (line->str, "-- "))
			break;
		pan_g_string_strstrip (line);
		if (line->len==0)
			continue;
		if (*line->str != '>') {
			g_free (body2);
			return;
		}
	}
	g_free (body2);

	g_ptr_array_add (setme_errors, g_strdup(
		_("ERROR: Message appears to have no new content.")));
	*setme_goodness = MAX (*setme_goodness, REFUSE);
}

static void
check_body (GPtrArray       * setme_errors,
	    GoodnessLevel   * setme_goodness,
	    const Article   * article,
	    const gchar     * body)
{
	GString * line;

	g_return_if_fail (setme_goodness != NULL);
	g_return_if_fail (setme_errors != NULL);
	g_return_if_fail (body != NULL);
	g_return_if_fail (article != NULL);

	line = g_string_sized_new (1024);
	check_empty (setme_errors, setme_goodness, body, line);
	check_all_quoted (setme_errors, setme_goodness, body, line, article);
	check_wide_body (setme_errors, setme_goodness, body, line);
	check_signature (setme_errors, setme_goodness, body, line);
	check_mostly_quoted (setme_errors, setme_goodness, body, line);
	g_string_free (line, TRUE);
}



static void
check_followup_to (GPtrArray        * setme_errors,
		   GoodnessLevel    * setme_goodness,
		   Server           * server,
		   const GPtrArray  * group_names)
{
	int i;

	g_return_if_fail (setme_errors != NULL);
	g_return_if_fail (setme_goodness != NULL);
	g_return_if_fail (server != NULL);
	g_return_if_fail (group_names != NULL);

	/* check to make sure all the groups exist */
	for (i=0; i<group_names->len; ++i)
	{
		const char * name = (char*) g_ptr_array_index (group_names, i);
		if (g_strcasecmp(name, "poster")
			&& !server_has_group (server, name))
		{
			g_ptr_array_add (setme_errors, g_strdup_printf(
				_("WARNING: Unknown group: \"%s\"."), name));
			*setme_goodness = MAX (*setme_goodness, WARN);
		}
	}

	/* warn if too many followup-to groups */
	if (group_names->len > 5)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Following-Up to many groups.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

static void
check_subject (GPtrArray * setme_errors,
	       GoodnessLevel * setme_goodness,
	       const gchar * subject)
{
	g_assert (setme_errors!=NULL);
	g_assert (setme_goodness!=NULL);

	if (!is_nonempty_string (subject))
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("ERROR: No subject specified.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
}

static void
check_groups (GPtrArray          * setme_errors,
	      GoodnessLevel      * setme_goodness,
	      Server             * server,
	      const GPtrArray    * group_names,
	      gboolean             followup_to_set)
{
	int i;

	g_return_if_fail (setme_errors!=NULL);
	g_return_if_fail (setme_goodness!=NULL);
	g_return_if_fail (server!=NULL);
	g_return_if_fail (group_names!=NULL);

	/* check to make sure all the groups exist */
	for (i=0; i<group_names->len; ++i)
	{
		const char * name = (char*) g_ptr_array_index (group_names, i);
		if (!server_has_group (server, name))
		{
			g_ptr_array_add (setme_errors, g_strdup_printf(
				_("WARNING: Unknown group: \"%s\"."), name));
			*setme_goodness = MAX (*setme_goodness, WARN);
		}
	}

	/* refuse if far too many groups */
	if (group_names->len > 12)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("ERROR: Posting to a very large number of groups.")));
		*setme_goodness = MAX (*setme_goodness, REFUSE);
	}
	/* warn if too many groups */
	else if (group_names->len > 5)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Posting to a large number of groups.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}


	/* warn if too many groups and no followup-to */
	if (group_names->len>2 && !followup_to_set)
	{
		g_ptr_array_add (setme_errors, g_strdup(
			_("WARNING: Crossposting without setting Followup-To.")));
		*setme_goodness = MAX (*setme_goodness, WARN);
	}
}

/**
 * @return the number of recipients
 */
static int
check_mail_addresses (GPtrArray      * setme_errors,
		      GoodnessLevel  * setme_goodness,
		      const char     * addresses_str)
{
	GPtrArray * a = gnksa_split_addresses (addresses_str);
	guint i;

	for (i=0; i!=a->len; ++i)
	{
		const gchar * from = (gchar*) g_ptr_array_index (a, i);
		const int val = gnksa_check_from (from);
		if (val != GNKSA_OK)
		{
			*setme_goodness = MAX (*setme_goodness, REFUSE);
			g_ptr_array_add (setme_errors, g_strdup_printf (
				_("ERROR: %s ``%s''"),
				gnksa_status_text(val), from));
		}
	}

	/* cleanup */
	i = a->len;
	pan_g_ptr_array_foreach (a, (GFunc)g_free, NULL);
	g_ptr_array_free (a, TRUE);
	return i;
}


void
check_article (const Article    * article,
               Server           * server,
               GPtrArray        * appendme_errors,
               GoodnessLevel    * setme_goodness)
{
	gchar * tmp;
	gchar * pch;
	int mail_qty;
	int group_qty;
	int author_qty;
	const gchar * cpch;
	gboolean followup_to_set;
	GoodnessLevel goodness = OKAY;

	/* sanity check */
	g_return_if_fail (article!=NULL);
	g_return_if_fail (appendme_errors!=NULL);
	g_return_if_fail (setme_goodness!=NULL);

	/* check the subject... */
	check_subject (appendme_errors, &goodness, article->subject);

	/* check the body... */
	tmp = article_get_body (article);
	check_body (appendme_errors, &goodness, article, tmp);
	g_free (tmp);

	/* check the optional followup-to... */
	followup_to_set = FALSE;
	cpch = article_get_header (article, HEADER_FOLLOWUP_TO);
	if (is_nonempty_string (cpch))
	{
		GPtrArray * groups = get_nntp_rcpts (cpch);

		followup_to_set = groups->len > 0;
		check_followup_to (appendme_errors,
		                   &goodness,
		                   server,
		                   groups);

		pan_g_ptr_array_foreach (groups, (GFunc)g_free, NULL);
		g_ptr_array_free (groups, TRUE);
	}

	/* check the groups... */
	group_qty = 0;
	cpch = article_get_header (article, HEADER_NEWSGROUPS);
	if (is_nonempty_string(cpch))
	{
		GPtrArray * groups = get_nntp_rcpts (cpch);

		group_qty = groups->len;
		check_groups (appendme_errors,
		              &goodness,
		              server,
		              groups,
		              followup_to_set);

		pan_g_ptr_array_foreach (groups, (GFunc)g_free, NULL);
		g_ptr_array_free (groups, TRUE);
	}

	/* check the author... */
	pch = article_get_author_str (article);
	author_qty = check_mail_addresses (appendme_errors,
	                                   &goodness,
	                                   pch);
	g_free (pch);
	if (author_qty!=1) {
		g_ptr_array_add (appendme_errors, g_strdup(
			_("ERROR: Must have one author in From: line")));
		goodness = MAX (goodness, REFUSE);
	}

	/* check the email recipients... */
	cpch = article_get_header (article, PAN_MAIL_TO);
	mail_qty = check_mail_addresses (appendme_errors, &goodness, cpch);

	/* one last error check */
	if (!group_qty && !mail_qty) {
		g_ptr_array_add (appendme_errors, g_strdup(
			_("ERROR: No recipients specified.")));
		goodness = MAX (goodness, REFUSE);
	}

	*setme_goodness = goodness;
}

void
check_article_and_prompt (const Article    * article,
                          Server           * server,
                          GPtrArray        * appendme_errors,
                          GoodnessLevel    * setme_goodness)
{
	check_article (article, server, appendme_errors, setme_goodness);

	/* post the warnings, if we have any */
	if (appendme_errors->len != 0)
	{
		guint i;
		GString * str;

		/* build the error message */
		str = g_string_new (NULL);
		for (i=0; i!=appendme_errors->len; ++i) {
			char * e = (char*)g_ptr_array_index(appendme_errors,i);
			g_string_append (str, e);
			g_string_append_c (str, '\n');
		}
		g_warning (str->str);

		/* either refuse, or ask the user if they're sure. */
		if (*setme_goodness==REFUSE)
		{
			pan_error_dialog (str->str);
		}
		else
		{
			guint button;
			const gchar *ques = appendme_errors->len > 1
				? _("Post anyway, despite these problems?")
				: _("Post anyway, despite this problems?");
			g_string_append (str, "\n");
			g_string_append (str, ques);
			button = gnome_dialog_run_and_close (GNOME_DIALOG(
					gnome_message_box_new (str->str,
						GNOME_MESSAGE_BOX_QUESTION,
						GNOME_STOCK_BUTTON_YES,
						GNOME_STOCK_BUTTON_NO,
						NULL)));
			if (button==0) /* yes, post anyway */
				*setme_goodness = OKAY;
		}

		/* cleanup */
		g_string_free (str, TRUE);
	}
}
