/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "acache.h"
#include "globals.h"
#include "gnksa.h"
#include "log.h"
#include "nntp.h"
#include "task.h"
#include "util.h"

#include "../libcamel/camel-mime-utils.h"
#include "../libcamel/broken-date-parser.h"

gboolean online_prompt = TRUE;

extern const char * sockwrite_err_msg;

extern const char * sockread_err_msg;

/**
***
**/

void
nntp_disconnect (PanSocket * sock)
{
	g_return_if_fail (sock!=NULL);

	pan_socket_putline (sock, "QUIT\r\n");
}

/**
***
**/


const gchar*
nntp_login (PanSocket   * sock,
            Server      * server)
{
	int ival;
	const gchar* buffer;

	g_return_val_if_fail (sock!=NULL, NULL);

	if (pan_socket_getline (sock, &buffer))
		return sockread_err_msg;

	/* attempt authentication.  This isn't implemented exactly as
	 * "Common NNTP Extensions" specifies, being that we didn't
	 * wait to be asked for it, but hopefully it will work.
	 */
	if (server->need_auth)
	{
		if (pan_socket_putline_va (sock, "AUTHINFO user %s\r\n", server->username))
			return sockwrite_err_msg;
		if (pan_socket_getline (sock, &buffer))
			return sockread_err_msg;
		if (atoi(buffer) == 381)
		{
			/* 381 Means: More authentication information needed */
			if (pan_socket_putline_va (sock, "AUTHINFO pass %s\r\n", server->password))
				return sockwrite_err_msg;
			if (pan_socket_getline (sock, &buffer))
				return sockread_err_msg;

			/* 281 Means: Authentication accepted */
			if (atoi(buffer) != 281) {
				log_add_va (_("Authentication for server \"%s\" unsuccessful."),
					server->name);
				return buffer;
			}
		}
		/* Else authentication not required or not understood */
	}

	/* find out if we have authorization to post */
	if (pan_socket_putline (sock, "MODE READER\r\n"))
		return sockwrite_err_msg;
	if (pan_socket_getline (sock, &buffer))
		return sockread_err_msg;
	ival = atoi(buffer);
	if (ival == 200)
		server->posting = TRUE;
	else if (ival == 201)
		server->posting = FALSE;
	else if (ival == 500) /* command not recognized; "mode reader" is an extension */
		server->posting = TRUE; /* a wild guess */
	else
		return buffer;

	return NULL;
}

/**
***
**/

gint
nntp_set_group (StatusItem   * status,
                PanSocket    * sock,
                const char   * group_name)
{
	const gchar *buffer = NULL;

	g_return_val_if_fail (sock!=NULL, TASK_FAIL_HOPELESS);
	g_return_val_if_fail (is_nonempty_string(group_name), TASK_FAIL_HOPELESS);

	/* write the request */
	if (pan_socket_putline_va (sock, "GROUP %s\r\n", group_name))
		return TASK_FAIL;

	/* get the result string */
	if (pan_socket_getline (sock, &buffer))
		return TASK_FAIL;

	/* interpret the result string */
	if (atoi(buffer)!=210 && atoi(buffer)!=211) {
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		gchar * err = g_strdup_printf (_("Unable to set group: server said %s"), tmp);
		if (status != NULL)
			status_item_emit_error (status, err);
		pan_error_dialog (err);
		g_free (err);
		g_free (tmp);
		return TASK_FAIL_HOPELESS;
	}

	return TASK_SUCCESS;
}

/**
***
**/

int
nntp_post (PanSocket     * sock,
           const gchar   * msg)
{
	const gchar * buffer;

	/* if we're in mute mode, don't post */
	if (Pan.mute) {
		g_message ("Mute: Your Message will not actually be posted.");
		printf ("\n\n\nYour message\n%s<end of message>\n\n\n", msg);
		fflush (NULL);
		return TASK_SUCCESS;
	}

	/* tell the server we want to post */
	if (pan_socket_putline (sock, "POST\r\n")) {
		pan_error_dialog (_("Error writing to socket during post."));
		return TASK_FAIL;
	}
	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during post."));
		return TASK_FAIL;
	}
	if (atoi(buffer) != 340) { /* 340 Means: Send Article To Be Posted */
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		gchar * err = g_strdup_printf (_("Error posting message.  Server sait: %s"), tmp);
		pan_error_dialog (err);
		g_free (err);
		g_free (tmp);
		return TASK_FAIL;
	}

	/* post the article */
	if (pan_socket_putline_va (sock, "%s\r\n.\r\n", msg)) {
		pan_error_dialog (_("Error writing to socket during post."));
		return TASK_FAIL;
	}
	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during post."));
		return TASK_FAIL;
	}
	if (atoi(buffer) != 240) { /* 240 Means: Article Posted OK */
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		gchar * err = g_strdup_printf (_("Error posting message.  Server sait: %s"), tmp);
		pan_error_dialog (err);
		g_free (err);
		g_free (tmp);
		return TASK_FAIL;
	}

	/* if we made it this far, we succeeded! */
	return TASK_SUCCESS;
}

/**
***
**/

gint
nntp_send_noop (PanSocket * sock)
{
	const char * buffer;

	/* the no-op is "mode reader", because it's a lightweight message */
	if (pan_socket_putline (sock, "MODE READER\r\n")) {
		pan_error_dialog (_("Error writing to socket during no-op."));
		return TASK_FAIL;
	}
	if (pan_socket_getline (sock, &buffer)) {
		pan_error_dialog (_("Error reading from socket during no-op."));
		return TASK_FAIL;
	}

	/* if we made it this far, we've succeeded. */
	return TASK_SUCCESS;
}

/**
***
**/

gint
nntp_article_download (StatusItem      * status,
                       PanSocket       * sock,
                       Group           * group,
                       Article         * article,
                       const gboolean  * abort,
                       gboolean          verbose)
{
	FILE * f;
	const gchar * buffer;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, TASK_FAIL);

	/* let the user know what we're doing */
	if (verbose)
		status_item_emit_status (status, _("Downloading article \"%s\""), article->subject);

	/**
	***  Tell the server to give us the headers
	**/

	if (*abort || pan_socket_putline_va (sock, "HEAD %lu\r\n", article->number)) {
		status_item_emit_error (status, sockwrite_err_msg);
		return TASK_FAIL;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}
	if (atoi(buffer) != 221) {
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		status_item_emit_error(status, _("NNTP \"HEAD\" command returned \"%s\"."), tmp);
		g_free (tmp);
		article_add_flag (article, STATE_ERROR);
		return TASK_FAIL_HOPELESS;
	}
	if (*abort || pan_socket_getline (sock,&buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}

	/**
	***  Walk through until we've gotten all the headers
	**/

	f = acache_open_header (article);
	while (strncmp(buffer, ".\r\n", 3) != 0)
	{
		gchar * pch = NULL;
		gchar * decoded;
		const gchar * name = NULL;

		/* check for exit... */
		if (*abort) {
			acache_close_header (article, f);
			acache_delete (article->message_id);
			return TASK_FAIL;
		}

		/* decode the header */
		decoded = header_decode_string (buffer);

		/* write to acache... */
		acache_putline_header (f, decoded);

		/* trim linefeed */
		if ((pch = strrchr (decoded, '\x0D')))
			*pch = '\0';

		/* pull out headers that we'd like to keep... */
		name = "Newsgroups: ";
		if (!strncmp (decoded, name, strlen(name)))
			article_set_newsgroups_chunk (article,
				group_chunk_string(article->group, decoded+strlen(name)));
		name = "Reply-To: ";
		if (!strncmp (decoded, name, strlen(name)))
			article_set_reply_to_chunk (article,
				group_chunk_string(article->group, decoded+strlen(name)));
		name = "Followup-To: ";
		if (!strncmp (decoded, name, strlen(name)))
			article_set_followup_to_chunk (article,
				group_chunk_string(article->group, decoded+strlen(name)));
		name = "Content-Transfer-Encoding: ";
		if (!strncmp (decoded, name, strlen(name)))
			article_set_encoding_chunk (article,
				group_chunk_string(article->group, decoded+strlen(name)));

		/* cleanup the iteration */
		g_free (decoded);
		decoded = NULL;

		/* get the next line... */
		if (*abort || pan_socket_getline (sock, &buffer)) {
			status_item_emit_error (status, sockread_err_msg);
			acache_close_header (article, f);
			acache_delete (article->message_id);
			return TASK_FAIL;
		}
	}
	acache_close_header (article, f);

	/**
	***  Tell the server to give us the body
	**/

	if (*abort || pan_socket_putline_va (sock, "BODY %lu\r\n", article->number)) {
		status_item_emit_error (status, sockwrite_err_msg);
		return TASK_FAIL;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}
	if (atoi(buffer) != 222) {
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		status_item_emit_error(status, _("NNTP \"BODY\" command returned \"%s\"."), tmp);
		g_free (tmp);
		article_add_flag (article, STATE_ERROR);
		return TASK_FAIL_HOPELESS;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}

	/**
	***  Walk through until we've gotten the entire body
	**/

	f = acache_open_body (article);
	if (verbose)
		status_item_emit_init_steps (status, article->linecount);
	while (strncmp(buffer, ".\r\n", 3))
	{
		if (*abort) {
			acache_close_body (article, f);
			acache_delete (article->message_id);
			return TASK_FAIL;
		}
		acache_putline_body (f, buffer);

		if (verbose)
			status_item_emit_next_step (status);

		if (*abort || pan_socket_getline (sock, &buffer)) {
			status_item_emit_error (status, sockread_err_msg);
			acache_close_body (article, f);
			acache_delete (article->message_id);
			return TASK_FAIL;
		}
	}
	acache_close_body (article, f);
	acache_file_checkin (article->message_id);

	if (verbose)
		status_item_emit_progress(status, 100);

	article_remove_flag (article, STATE_DOWNLOAD_FLAGGED);

	return TASK_SUCCESS;
}

/**
***
**/

gint
nntp_get_group_info (StatusItem       * status,
                     PanSocket        * sock,
                     const char       * group_name,
                     gulong           * article_qty,
                     gulong           * low_num,
                     gulong           * high_num,
                     const gboolean   * abort)
{
	const gchar *buffer = NULL;

	/* change to the right group */
	if (*abort || pan_socket_putline_va (sock, "GROUP %s\r\n", group_name)) {
		return TASK_FAIL;
	}
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}
	if (strtol(buffer, NULL, 10) != 211) {
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		gchar * err = g_strdup_printf (_("NNTP \"GROUP %s\" command returned \"%s\"."), group_name, tmp);
		status_item_emit_error (status, err);
		g_free (err);
		g_free (tmp);
		return TASK_FAIL_HOPELESS;
	}
	sscanf (buffer, "%*d %lu %lu %lu", article_qty, low_num, high_num);
	return TASK_SUCCESS;
}

/**
***
**/

gint
nntp_articlelist_download (StatusItem       * status,
                           PanSocket        * sock,
                           Group            * group,
                           gulong             low,
                           gulong             high,
                           const gboolean   * abort,
                           const char       * progress_fmt,
                           GPtrArray        * setme)
{
	Article * article;
	const char * buffer;
	gulong total;
	gulong first;
	gulong last;
	gboolean read_status = 0;
	GString * work;
	gint result;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, TASK_FAIL_HOPELESS);
	g_return_val_if_fail (group!=NULL, TASK_FAIL_HOPELESS);

	/**
	***  Get up-to-date information about this group
	**/

	total = first = last = 0;
	result = nntp_get_group_info (status,
	                              sock,
	                              group->name,
	                              &total,
	                              &first,
	                              &last,
	                              abort);
	if (result != TASK_SUCCESS)
		return result;


	/**
	***  If no articles to get, then we're already done
	**/

	if (total == 0) {
		const gchar * n = group_get_readable_name (group);
		status_item_emit_status (status, _("No articles found for group \"%s\""), n);
		return TASK_SUCCESS;
	}

	/**
	***  Tell the server that we want a series of article headers...
	**/

	status_item_emit_status (status, _("Getting %d article headers..."), high-low);
	if (*abort || pan_socket_putline_va (sock, "XOVER %lu-%lu\r\n", low, high)) {
		status_item_emit_error (status, sockwrite_err_msg);
		return TASK_FAIL;
	}
	status_item_emit_status (status, _("Waiting for Response from Server..."));
	if (*abort || pan_socket_getline (sock, &buffer)) {
		status_item_emit_error (status, sockread_err_msg);
		return TASK_FAIL;
	}
	if (atoi(buffer) != 224) {
		gchar * tmp = pan_substitute (buffer, "\r\n", "");
		gchar * err = g_strdup_printf (_("Can't get headers: NNTP XOVER command returned \"%s\"."), tmp);
		status_item_emit_error (status, err);
		g_free (err);
		g_free (tmp);
		return TASK_FAIL_HOPELESS;
	}

	/**
	***  Walk through all the articles headers that the server spits back
	**/

       	work = g_string_new (NULL);
	while (!*abort && !((read_status = pan_socket_getline (sock, &buffer))))
	{
		gint part = 0;
	        gint parts = 0;
		const gchar * pch;
		const gchar * s;
		gchar * tmp;

		/* handle end-of-list */
		if (!strncmp(buffer, ".\r\n", 3))
			break;

		/* create the article data */
		pch = buffer;
		article = article_new (group);
		group->articles_dirty = TRUE;

		/* article number */
		article->number = get_next_token_ulong (pch, '\t', &pch);

		/* subject */
		get_next_token_g_str (pch, '\t', &pch, work);
		tmp = header_decode_string (work->str);
		article_set_string_field (article, &article->subject, tmp);
		g_free (tmp);

		/* author */
		get_next_token_g_str (pch, '\t', &pch, work);
		tmp = header_decode_string (work->str);
		article_set_author_from_header (article, tmp);
		g_free (tmp);

		/* date */
		get_next_token_g_str (pch, '\t', &pch, work);
		article->date = header_decode_date (work->str, NULL);

		/* message id */
		get_next_token_g_str (pch, '\t', &pch, work);
		article_set_string_field (article, &article->message_id, work->str);

		/* references */
		get_next_token_g_str (pch, '\t', &pch, work);
		article_set_string_field (article, &article->references, work->str);

		/* unused field */
		get_next_token_g_str (pch, '\t', &pch, work);

		/* linecount */
		article->linecount = get_next_token_int(pch, '\t', &pch);

		/* crosspost */
		tmp = get_next_token_str (pch, '\t', &pch);
		if (tmp==NULL || strncmp (tmp, "Xref:", 5))
			article->crosspost_qty = 1;
		else {
			gchar * foo;
			article->crosspost_qty = 0;
			for (foo=tmp+5; *foo; ++foo)
				if (*foo==':')
					++article->crosspost_qty;
		}
		g_free (tmp);

		/**
		***  Validity Checking
		**/

		if (!is_nonempty_string(article->subject)) {
			gchar * s = _("Corrupt header skipped - no subject");
			g_message (s);
			status_item_emit_error (status, "%s: %s", s, buffer);
			pan_object_unref (PAN_OBJECT(article));
			continue;
		}
		if (!is_nonempty_string(article->message_id)) {
			gchar * s = _("Corrupt header skipped - no msg id");
			g_message (s);
			status_item_emit_error (status, "%s: %s", s, buffer);
			pan_object_unref (PAN_OBJECT(article));
			continue;
		}
		if (!is_nonempty_string(article->message_id)) {
			gchar * s = _("Corrupt header skipped - no date");
			g_message (s);
			status_item_emit_error (status, "%s: %s", s, buffer);
			pan_object_unref (PAN_OBJECT(article));
			continue;
		}
		tmp = article_get_author_str (article);
		if (!is_nonempty_string(tmp)) {
			gchar * s = _("Corrupt header skipped - no author");
			g_message (s);
			status_item_emit_error (status, "%s: %s", s, buffer);
			pan_object_unref (PAN_OBJECT(article));
			g_free (tmp);
			continue;
		}
		g_free (tmp);


		/* Look for the (n/N) or [n/N] construct in subject lines,
		 * starting at the end of the string and working backwards */
		part = 0;
		parts = 0;
		s = article->subject;
		pch = s + strlen(s) - 1;
		while (pch != s)
		{
			/* find the ']' of [n/N] */
			--pch;
			if ((pch[1]!=')' && pch[1]!=']') || !isdigit((int)*pch))
				continue;

			/* find the '/' of [n/N] */
			while (s!=pch && isdigit((int)*pch))
				--pch;
			if (s==pch || (*pch!='/' && *pch!='|'))
				continue;

			/* N -> parts */
			parts = atoi (pch+1);
			--pch;

			/* find the '[' of [n/N] */
			while (s!=pch && isdigit((int)*pch))
				--pch;
			if (s==pch || (*pch!='(' && *pch!='[')) {
				parts = 0;
				continue;
			}

			/* n -> part */
			part = atoi (pch+1);
			break;
		}

		/* if not a multipart yet, AND if it's a big message, AND
		   it's either in one of the pictures/fan/sex groups or it
		   has commonly-used image names in the subject, guess it's
		   a single-part binary */
		if (!parts
			&& article->linecount>400
			&& (((strstr(group->name,"binaries")
					|| strstr(group->name,"fan")
				        || strstr(group->name,"mag")
					|| strstr(group->name,"sex")))
				|| ((strstr(s,".jpg") || strstr(s,".JPG")
					|| strstr(s,".jpeg") || strstr(s,".JPEG")
					|| strstr(s,".gif") || strstr(s,".GIF")
					|| strstr(s,".tiff") || strstr(s,".TIFF")))))
			part = parts = 1;

		/* Verify Multipart info */
		if ((parts>=1) && (part<=parts)) {
			/* yay, we found one! */
			article->parts = parts;
			article->part = part;
		}
		else {
			article->parts = 0;
			article->part = 0;
		}

		/* Add the article to the article list */
		g_ptr_array_add (setme, article);

		/* update the count & progress ui */
		status_item_emit_next_step (status);
		if (!(setme->len % 29))
			status_item_emit_status (status, progress_fmt, setme->len, (high-low));
	}
       	g_string_free (work, TRUE);
	if (read_status!=0)
		status_item_emit_error (status, sockread_err_msg);
	if (read_status!=0 || *abort)
		return TASK_FAIL;

	/* if we've made it this far, we succeeded! */
	return TASK_SUCCESS;
}

/**
***
**/

gint
nntp_download_bodies (StatusItem       * status,
                      Group            * group,
                      PanSocket        * sock,
                      const gboolean   * abort,
                      const GPtrArray  * const articles,
                      gint             * index)
{
	gint result;

	/* sanity checks */
	g_return_val_if_fail (group!=NULL, TASK_FAIL_HOPELESS);
	g_return_val_if_fail (articles!=NULL, TASK_FAIL_HOPELESS);

	/* change to the group we want */
	if (*abort)
		return TASK_FAIL;
	result = nntp_set_group (status, sock, group->name);
	if (result != TASK_SUCCESS) {
		const gchar * g = group_get_readable_name (group);
		status_item_emit_error (status, _("Couldn't select group \"%s\""), g);
		return result;
	}

	/* download the articles */
	for (; *index<articles->len && !*abort; ++(*index))
	{
		Article * article = ARTICLE(g_ptr_array_index(articles,*index));
		gboolean need_download;

		/* download if we have to.  Abort if the download failed but
		 * might succeed later.  Hopeless downloads silently fall
		 * through; it's pointless to postpone the task for them. */
		need_download = !acache_body_exists (article->message_id) ||
		                !acache_header_exists (article->message_id);

		if (need_download)
		{
			result = nntp_article_download (status, sock, group, article, abort, FALSE);
		}
		else
		{
			result = TASK_SUCCESS;

			/* since it's cached locally, we can unflag */
			article_remove_flag (article, STATE_DOWNLOAD_FLAGGED);
		}

		switch (result)
		{
			case TASK_FAIL:
				status_item_emit_status (status, _("Downloading article \"%s\" failed."), article->subject);
				return result;
			case TASK_SUCCESS:
				status_item_emit_next_step (status);
				status_item_emit_status (status, _("Getting article %d of %d"), 1+*index, articles->len);
				break;
			default:
				pan_warn_if_reached ();
		}
	}

	return TASK_SUCCESS;
}
