/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999, 2000  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>

#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>

#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>

#if defined(__FreeBSD__) || defined(__svr4__) || defined(__OpenBSD__)
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif

#ifdef AF_NEEDS_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>

#include "debug.h"
#include "globals.h"
#include "gnksa.h"
#include "util.h"
#include "prefs.h"




void
pan_error_dialog (const char *format, ...)
{
	va_list args;
	char *str = NULL;
	g_return_if_fail (format != NULL);
	va_start (args, format);
	str = g_strdup_vprintf (format, args);
	va_end (args);
	g_warning (str);
	pan_lock();
	gnome_error_dialog (str);
	pan_unlock();
	g_free (str);
}



static pthread_t has_lock_thr = 0;

void
pan_lock_unconditional_from (const gchar * file, int line)
{
	pthread_t thr = pthread_self();
	if (thr == has_lock_thr)
		debug3 (DEBUG_LOCK,"thread %lu attempted double lock from %s %d!", thr, file, line);
	gdk_threads_enter();
	has_lock_thr = thr;
	debug3 (DEBUG_LOCK,"thread %lu entered gdk_threads from %s %d", thr, file, line);
}

void
pan_lock_from (const gchar * file, int line)
{
	pthread_t thr = pthread_self();

	if (thr==Pan.main_t && !has_lock_thr)
	{
		debug3 (DEBUG_LOCK,"mainthread %lu attempted lock from %s %d", thr, file, line);
	}
	else
	{
		pan_lock_unconditional_from (file, line);
	}
}

void
pan_unlock_unconditional_from (const gchar * file, int line)
{
	has_lock_thr = (pthread_t)0;
	gdk_threads_leave();
	debug3 (DEBUG_LOCK,"thread %lu left gdk_threads from %s %d", pthread_self(), file, line);
}

void
pan_unlock_from (const gchar* file, int line)
{
	pthread_t thr = pthread_self();

	if (thr == Pan.main_t)
	{
		debug3 (DEBUG_LOCK,"mainthread %lu attempted unlock from %s %d", thr, file, line);
	}
	else
	{
		pan_unlock_unconditional_from (file, line);
	}
}

/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
void
directory_check (const char *pathname)
{
	GString * dir = g_string_new (NULL);
	struct stat st;
	int len;
	int i;

	g_return_if_fail (pathname!=NULL);

	for (len=strlen(pathname), i=1; i<=len; ++i)
	{
		if (pathname[i] == '/' || pathname[i] == '\0')
		{
			int statval;

			/* check the directory */
			g_string_assign (dir, pathname);
			g_string_truncate (dir, i);
			statval = stat (dir->str, &st);

			/* error checking */
			if (statval != 0)
			{
				int ret;
				ret = mkdir (dir->str, S_IRUSR | S_IWUSR | S_IXUSR);
				if (ret == -1) {
					pan_error_dialog (_("Couldn't create directory: "),
							  pathname, g_strerror (errno));
					break;
				}
			}
		}
	}

	g_string_free (dir, TRUE);
}


int
lower_bound (const void         * key,
             const void         * base,
             size_t               n,
             size_t               size,
             int                  (*compare)(const void *, const void *),
             gboolean           * exact_match )
{
	int low = 0;
	int high = n-1;

	while (low<=high)
	{
		const int mid = (low+high)/2;
		const void * checkme = (void*)(((char*)(base))+(mid*size));
		const int comp = (*compare)(key,checkme);

		if (comp>0) low = mid+1;
		else if (comp<0 ) high = mid-1;
		else {
			if ( exact_match!=NULL )
				*exact_match = TRUE;
			return mid;
		}
	}

	if (exact_match!=NULL)
		*exact_match = FALSE;

	return low;
}



GArray*
read_file (const char* filename)
{
	GArray* array = NULL;
	FILE* fp;

	/* get length of file... */
	size_t len = get_filesize(filename);
	if (len<=0)
		return NULL;

	/* open the file... */
	fp = fopen ( filename, "r" );
	if (!fp)
		return NULL;

	/* read the file... */
	array = g_array_new (TRUE, FALSE, 1);
	g_array_set_size (array, len);
	array->len = fread (array->data, 1, len, fp);
	fclose (fp);
	return array;
}

void
commatize_ulong (gulong    num,
		 char    * setme)
{
	char buf[16], *src=buf;
	int len;
	int i;

	sprintf ( buf, "%ld", num );
	len = strlen ( buf );
	i = len % 3;
	for ( ;; ) {
		while ( *src && i-- )
			*setme++ = *src++;
		if ( !*src ) {
			*setme = '\0';
			return;
		}
		else if ( src!=buf )
		   *setme++ = ',';
		i = 3;
	}

	pan_warn_if_reached ();
}

/**
 * Use stat() to get and then return the size of the file.
 */
size_t
get_filesize (const char* filename)
{
	struct stat buf;
	if ( !stat(filename, &buf) )
		return buf.st_size;

	return -1;
}




/**
 * Replaces the search string inside original with the replace string.
 * This should be safe with overlapping elements of text, though I haven't
 * not tested it with elaborately cruel cases.
 */
char*
pan_substitute (const char *original,
                const char* search,
                const char* replace)
{
	size_t slen;		/* length of search */
	size_t rlen;		/* length of replace */
	size_t tlen;		/* length of target (predicted) */
	int i;
	const char * o;
	const char * pchar;
	char * t;
	char * retval = NULL;

	g_return_val_if_fail (original!=NULL, NULL);
	g_return_val_if_fail (*original!='\0', NULL);
	g_return_val_if_fail (search!=NULL, NULL);
	g_return_val_if_fail (*search!='\0', NULL);
	g_return_val_if_fail (replace!=NULL, NULL);

	slen = strlen (search);
	rlen = strlen (replace);

	/* calculate the length */

	i = 0;
	tlen = 0;
	pchar = original;
	while ((pchar = strstr (pchar, search))) {
		i++;
		pchar += slen;
	}
	tlen = strlen(original) + i*(rlen - slen);

	/**
	***  Make the substitution.
	**/

	o = original;
	t = retval = g_malloc(tlen + 1);
	while ((pchar = strstr (o, search))) {
		(void) memcpy (t, o, (size_t)(pchar-o));
		t += pchar-o;
		(void) memcpy (t, replace, (size_t)rlen);
		t += rlen;
		o = pchar + slen;
	}
	(void) strcpy ( t, o );

	pan_warn_if_fail (strlen(retval)==tlen);

	return retval;
}

char *
pan_make_temp (FILE ** setme_fp)
{
	char * retval = NULL;
	char * fname = g_strdup_printf ("/%s/pan%uXXXXXX", temp_dir, (guint)getpid());
	int fd = mkstemp (fname);

	if (fd == -1)
		g_free (fname);
	else {
		retval = fname;
		*setme_fp = fdopen (fd, "w+");
	}

	return retval;
}

void
pan_clean_temp_dir (void)
{
	GString *filename = g_string_new (NULL);
	char *dirpath = g_strdup_printf ("/%s", temp_dir); 
	char *prefix = g_strdup_printf ("pan%u", (unsigned int)getpid ());
	gint prefix_l = strlen (prefix);
	DIR *dir_p = opendir(dirpath);
	struct dirent *dirent_p;

	while ((dirent_p = readdir(dir_p)))
	{
		if (!strncmp (prefix, dirent_p->d_name, prefix_l))
		{
			g_string_sprintf (filename,
					  "/%s/%s", temp_dir, dirent_p->d_name);
			unlink (filename->str);
		}	
	}
	closedir (dir_p);
	g_free (prefix);
	g_free (dirpath);
	g_string_free (filename, TRUE);
}

gboolean
file_exists (const char* filename)
{
	return g_file_exists (filename);
}

char*
get_host_name (void)
{
	char *ptr;
	char hostname[256] = { '\0' };

	if ((gethostname(hostname, sizeof(hostname))) == -1) {
		if ((ptr = getenv("HOST")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else if ((ptr = getenv("HOSTNAME")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else
			hostname[0] = '\0';
	}
	hostname[255] = '\0';
	ptr = strtok (hostname, ".");

	return g_strdup (ptr);
}

/* Returns the fully qualified domain name */
char*
get_fqdn (const char * host)
{
	char name[512] = { '\0' };
	char line[1025];
	char fqdn[1024] = { '\0' };
	char * cp;
	char * domain = NULL;
	FILE * inf;
	struct hostent * hp;
	struct in_addr in;

	if (host) {
		if (strchr(host, '.'))
			return g_strdup(host);
		strncpy(name, host, MAXHOSTNAMELEN);
	}
	else {
		if (gethostname(name, MAXHOSTNAMELEN))
			return NULL;
	}

	if (isdigit((int)*name)) {
		in.s_addr = inet_addr (name);
		if ((hp = gethostbyaddr((char *) &in.s_addr, 4, AF_INET)))
			in.s_addr = (*hp->h_addr);
		return g_strdup((hp && strchr(hp->h_name, '.') ? hp->h_name : (char *) inet_ntoa(in)));
	}
	if ((hp = gethostbyname(name)) && !strchr(hp->h_name, '.'))
		if ((hp = gethostbyaddr(hp->h_addr, hp->h_length, hp->h_addrtype)))
			in.s_addr = (*hp->h_addr);

	sprintf(fqdn, "%s", hp ? strchr(hp->h_name, '.')
		? hp->h_name : (char *) inet_ntoa(in)
		: "");
	if (!*fqdn || (fqdn[strlen(fqdn) - 1] <= '9')) {
		*fqdn = '\0';
		inf = fopen("/etc/resolv.conf", "r");
		if (inf) {
			while (fgets(line, 1024, inf)) {
				line[1024] = '\0';
				g_strchug(line);
				g_strchomp(line);
				if (strncmp(line, "domain ", 7) == 0) {
					domain = line + 7;
					break;
				}
				if (strncmp(line, "search ", 7) == 0) {
					domain = line + 7;
					cp = strchr(domain, ' ');
					if (cp)
						*cp = '\0';
					break;
				}
			}
			if (domain)
				sprintf(fqdn, "%s.%s", name, domain);
			fclose(inf);
		}
	}
	return g_strdup (fqdn);
}


/* null-safe strncmp.  NULLs < strings. */
int
pan_strncmp (const char * a,
             const char * b,
             int          len)
{
	int retval;

	if (!a && !b)
		retval = 0;
	else if (!a)
		retval = -1;
	else if (!b)
		retval = 1;
	else
		retval = strncmp (a, b, len);

	return retval;
}

/* null-safe strcmp.  NULLs < strings. */
int 
pan_strcmp (const char * a,
            const char* b)
{
	int retval;

	if (!a && !b)
		retval = 0;
	else if (!a)
		retval = -1;
	else if (!b)
		retval = 1;
	else
		retval = strcmp (a, b);

	return retval;
}

void
array_shrink (gpointer   array,
              size_t     index,
              size_t     element_size,
              size_t     element_qty)
{
	g_return_if_fail (array!=NULL);
	g_return_if_fail (element_size!=0);
	g_return_if_fail (index>=0);
	g_return_if_fail (index<=element_qty);

	if (index != element_qty-1)
	{
		g_memmove (
			((char*)array) + element_size * index,
			((char*)array) + element_size * (index + 1),
			element_size * (element_qty - index -1));
	}
}

void
replace_gstr (char   ** target_gfree_old,
              char    * assign_from_me)
{
	char * gfree_old;

 	g_return_if_fail (target_gfree_old != NULL);

	gfree_old = *target_gfree_old;
	*target_gfree_old = assign_from_me;
	g_free (gfree_old);
}

/****
*****  LINE WRAPPING
****/

static void
get_leader (const char * line,
            GString    * setme)
{
	g_string_truncate (setme, 0);
	for (; is_nonempty_string(line) && (*line=='>' || isspace((int)*line)); ++line)
		g_string_append_c (setme, *line);
}

typedef struct
{
	char * leader;
	char * content;
}
Paragraph;

static void
paragraph_free (Paragraph * p)
{
	g_free (p->leader);
	g_free (p->content);
	g_free (p);
}

/**
 * FIXME: doesn't handle paragraphs whose first lines are indented
 */
static GPtrArray*
get_paragraphs (const char * body)
{
	int i = 0;
	char ** lines = g_strsplit (body, "\n", -1);
	GString * leader = g_string_new (NULL);
	GString * content = g_string_new (NULL);
	GString * last_leader = g_string_new (NULL);
	GPtrArray * paragraphs = g_ptr_array_new ();

	for (i=0; ; ++i)
	{
		char * line = NULL;
		get_leader (lines[i], leader);

		if (lines[i])
			line = lines[i] + leader->len;

		if (content->len /* something to keep... */
		    && (!line /* eof forces paragraph break */
		        || !*line /* empty line forces paragraph break */
			|| (line && !isalpha((int)*line)) /* source code? */
			|| pan_strcmp(leader->str,last_leader->str) /*break*/
		       )
		    )
		{
			Paragraph * p = g_new0 (Paragraph, 1);
			p->content = g_strdup (content->str);
			p->leader = g_strdup (last_leader->str);
			g_ptr_array_add (paragraphs, p);
			g_string_truncate (content, 0);
		}

		if (line) /* don't process a non-line */
		{
			if (!*line /* empty line forces paragraph break */
				|| !isalpha((int)*line)) /* source code */
			{
				Paragraph * p = g_new0 (Paragraph, 1);
				p->content = g_strdup (line);
				p->leader = g_strdup (leader->str);
				g_ptr_array_add (paragraphs, p);
			}
			else /* augment paragraph */
			{
				if (content->len)
					g_string_append_c (content, ' ');
				g_string_append (content, line);
			}
		}

		/* prepare for next iteration */
		g_string_assign (last_leader, leader->str);

		/* loop past the last line, to get even the last paragraph */
		if (!line)
			break;
	}

	/* make the paragraphs into one continuous line each */
	for (i=0; i<paragraphs->len; ++i)
	{
		Paragraph *p = (Paragraph*) g_ptr_array_index (paragraphs, i);
		char* pch;
		for (pch=p->content; *pch; ++pch)
		{
			/* if end of line, remove those characters */
			if (*pch=='\n')
			{
				/* also remove any trailing spaces */
				char * it = pch;
				while (it!=p->content && isspace((int)it[-1]))
					--it;
				g_memmove (it, pch+1, strlen(pch+1)+1); /* +1 for '\0' */
			}
		}
	}

	/* cleanup */
	g_string_free (content, TRUE);
	g_string_free (leader, TRUE);
	g_string_free (last_leader, TRUE);
	g_strfreev (lines);

	return paragraphs;
}

static char*
wrap_long_lines (char   * string,
                 gint     maxlen)
{
	char * linefeed_here;
	char * pch;
	char * line_start;


	/* walk through the entire string */
	linefeed_here = NULL;
	for (line_start=pch=string; *pch; )
	{
		/* a linefeed could go here; remember this space */
		if (isspace((int)*pch) || *pch=='\n')
			linefeed_here = pch;

		/* line's too long; add a linefeed if we can */
		if (pch-line_start>=maxlen && linefeed_here!=NULL)
		{
			*linefeed_here = '\n';
			pch = line_start = linefeed_here + 1;
			linefeed_here = NULL;
		}
		else ++pch;
	}

	return string;
}

static void
fill_paragraph (const Paragraph   * p,
                GPtrArray         * setme)
{
	GString * str = g_string_new (NULL);
	gchar * pch = g_strconcat (p->content, "\n", NULL);
	const gchar * march = pch;
	const int wrap_len = wrap_column - strlen(p->leader);

	wrap_long_lines (pch, wrap_len);
	while ((get_next_token_g_str (march, '\n', &march, str)))
		g_ptr_array_add (setme, g_strconcat(p->leader, str->str, NULL));

	g_free (pch);
	g_string_free (str, TRUE);
}

char*
fill_body (const char * body)
{
	GPtrArray* array = g_ptr_array_new ();
	GPtrArray* paragraphs;
	gchar * new_body;
	gchar * tmp_body;
	gchar * sig;

	/* sanity checks */
	g_return_val_if_fail (is_nonempty_string(body), NULL);

	/* get a temp copy of the body -- we don't wrap the signature. */
	tmp_body = pan_substitute (body, "\r", "");
	sig = strstr (tmp_body, "\n-- \n");
	if (sig != NULL)
		*sig = '\0';

	/* fill the paragraphs */
       	paragraphs = get_paragraphs (tmp_body);
	pan_g_ptr_array_foreach (paragraphs, (GFunc)fill_paragraph, array);

	/* make a single string of all filled lines */
	g_ptr_array_add (array, NULL);
	new_body = g_strjoinv ("\n", (char**)array->pdata);

	/* if we had a sig, put it back in */
	if (sig != NULL) {
		*sig = '\n';
		replace_gstr (&new_body, g_strjoin("\n",new_body, sig, NULL));
	}

	/* cleanup */
	pan_g_ptr_array_foreach (paragraphs, (GFunc)paragraph_free, NULL);
	g_ptr_array_free (paragraphs, TRUE);
	pan_g_ptr_array_foreach (array, (GFunc)g_free, NULL);
	g_ptr_array_free (array, TRUE);
	g_free (tmp_body);

	return new_body;
}

/****
*****  PAN GLIB UTILITIES
****/

void
pan_g_ptr_array_append (GPtrArray   * target,
                        gpointer    * source_ptr,
                        guint         source_qty)
{
	const guint old_len = target->len;
	g_ptr_array_set_size (target, old_len+source_qty);
	memcpy (target->pdata+old_len, source_ptr, sizeof(gpointer)*source_qty);
}

void
pan_g_ptr_array_assign  (GPtrArray  * target,
                         gpointer   * source_ptr,
                         guint        source_qty)
{
	g_ptr_array_set_size (target, source_qty);
	memcpy (target->pdata, source_ptr, sizeof(gpointer)*source_qty);
}


void
pan_g_ptr_array_reserve (GPtrArray  * a,
                         int          n)
{
	int len;

	g_return_if_fail (a!=NULL);

	len = a->len;
	g_ptr_array_set_size (a, MAX(len, n));
	a->len = len;
}

void
pan_g_ptr_array_insert (GPtrArray   * a,
                        gpointer      ptr,
                        int           index)
{
        g_return_if_fail (a!=NULL);

        if (index<0 || index>=a->len)
	{
                g_ptr_array_add (a, ptr);
	}
        else
        {
		pan_g_ptr_array_reserve (a, a->len+1);
                g_memmove (&a->pdata[index+1],
                           &a->pdata[index],
                           sizeof(gpointer)*(a->len-index));
                a->pdata[index] = ptr;
		++a->len;
        }
}

void
pan_g_ptr_array_foreach (GPtrArray   * a,
                         GFunc         func,
                         gpointer      user_data)
{
	int i;

	g_return_if_fail (a!=NULL);

	for (i=0; i<a->len; ++i)
	{
		gpointer call_data = g_ptr_array_index (a, i);
		(*func)(call_data, user_data);
	}
}

void
pan_g_string_replace (GString       * string,
                      const char    * search,
		      const char    * replace)
{
	const gchar * pch;
	size_t slen;
	size_t rlen;
	size_t offset;

	g_return_if_fail (string != NULL);
	g_return_if_fail (is_nonempty_string(search));
	g_return_if_fail (replace != NULL);

	slen = strlen (search);
	rlen = strlen (replace);

	offset = 0;
	while ((pch = strstr (string->str + offset, search)) && *pch) {
		const gint pos = pch - string->str;
		g_string_erase (string, pos, slen);
		g_string_insert (string, pos, replace);
		offset = pos + rlen;
	}
}

void
pan_g_string_strstrip (GString * string)
{
	g_return_if_fail (string != NULL);

	if (string->len != 0)
	{
		g_strstrip (string->str);
		string->len = strlen (string->str);
	}
}

/****
*****  TOKENIZING
****/

glong
get_next_token_long (const char   * pch,
                     const char     delimiter,
                     const char  ** setme_next_token)
{
	glong retval = 0;

	if (is_nonempty_string(pch))
	{
		retval = strtol (pch, NULL, 10);

		while (*pch && *pch!=delimiter)
			++pch;
		if (*pch==delimiter)
			++pch; /* skip past delimiter */
		if (setme_next_token)
			*setme_next_token = pch;
	}

	return retval;
}

gulong
get_next_token_ulong (const char   * pch,
                      const char     delimiter,
                      const char  ** setme_next_token)
{
	gulong retval = 0;

	if (is_nonempty_string(pch))
	{
		retval = strtoul (pch, NULL, 10);

		while (*pch && *pch!=delimiter)
			++pch;
		if (*pch==delimiter)
			++pch; /* skip past delimiter */
		if (setme_next_token)
			*setme_next_token = pch;
	}

	return retval;
}

int
get_next_token_int (const char   * pch,
                    const char     delimiter,
                    const char  ** setme_next_token)
{
	int retval = 0;

	if (is_nonempty_string(pch))
	{
		retval = atoi (pch);
		while (*pch && *pch!=delimiter)
			++pch;
		if (*pch==delimiter)
			++pch; /* skip past delimiter */
		if (setme_next_token)
			*setme_next_token = pch;
	}

	return retval;
}

char*
get_next_token_str (const char   * pch,
                    const char     delimiter,
                    const char  ** setme_next_token)
{
	char* retval = NULL;
	register const char* end = pch;

	while (is_nonempty_string(end) && *end!=delimiter)
		++end;

	if (end != pch)
	{
		int len = end - pch;
		retval = g_memdup (pch, 1+len);
		retval[len] = '\0';
	}

	if (setme_next_token)
		*setme_next_token = *end ? end+1 : end;

	return retval;
}

gboolean
get_next_token_g_str (const char   * pch,
                      const char     delimiter,
                      const char  ** setme_next_token,
                      GString      * setme)
{
	register const char* end = pch;
	gboolean retval;

	while (is_nonempty_string(end) && *end!=delimiter)
		++end;

	if (end != pch)
	{
		retval = TRUE;
		g_string_sprintf (setme, "%*.*s", end-pch, end-pch, pch);
	}
	else
	{
		retval = *end==delimiter;
		g_string_assign (setme, "");
	}

	if (setme_next_token)
		*setme_next_token = *end ? end+1 : end;

	return retval;
}

/****
*****
****/

#define DB_FORMAT_LAST_CHANGED_IN_VERSION "0.8.1 beta 3"
#define DB_FORMAT_VERSION 2

void
check_and_warn_if_version_change (void)
{
	int old_db_version =
		gnome_config_get_int ("/Pan/General/DBFormatVersion=0");
	gchar * username = gnome_config_get_string ("/Pan/User/Full_Name");
	const gboolean new_user = !is_nonempty_string (username);

	/* 0.8.1 beta 1 used a different key for finding db format version */
	if (old_db_version < 1)
	{
		gchar * s = gnome_config_get_string ("/Pan/General/AppVersion");
		if (!pan_strcmp(s, "0.8.1beta1"))
			old_db_version = 1;
		g_free (s);
	}

	/* do something if the versions differ... */
	if (old_db_version != DB_FORMAT_VERSION)
	{
	       	/* not a new user, let them know what's happening... */
		if (!new_user)
		{
			gchar * msg = g_strdup_printf (
_("Thanks for upgrading to Pan %s!"
"\n"
"\nThe database formats changed in version %s of Pan."
"\nYou need to delete your Pan database files (located in the directory"
"\n``.pan/data'' in your home directory, by default) before using this"
"\nversion of Pan."
"\n"
"\nFor more information, see the installation section of the"
"\nPan FAQ at <http://www.superpimp.org/faq.html>."),
				VERSION,
				DB_FORMAT_LAST_CHANGED_IN_VERSION);
			GtkWidget * dialog = gnome_message_box_new (msg,
				GNOME_MESSAGE_BOX_INFO,
				GNOME_STOCK_BUTTON_CANCEL,
				GNOME_STOCK_BUTTON_NEXT,
				NULL);
			int button = gnome_dialog_run_and_close (
				GNOME_DIALOG(dialog));
			if (button == 0)
				exit (0);
			g_free (msg);
		}

		/* update the config so that the user won't be warned
		   the next time they run Pan. */
		gnome_config_set_int (
			"/Pan/General/DBFormatVersion", DB_FORMAT_VERSION);
	}

	g_free (username);
}

/****
*****
****/

static gboolean
badchar (gchar c)
{
	return c==' '
		|| c==','
		|| c=='('
		|| c==')'
		|| c=='\0'
		|| c=='\n'
		|| c=='<'
		|| c=='>';
}

static const gchar*
end_of_url (const gchar * pch)
{
	while (!badchar(*pch))
		++pch;
	if (pch[-1] == '.')
		--pch;
	return pch;
}


gchar*
linkify_text (const char * text)
{
	GString * out = g_string_new (NULL);
        const gchar * c;

	/* walk through the text */
	c = text;
	while (*c)
	{
		/* pass over already-existing anchors */
		if (!g_strncasecmp (c, "<A", 2))
		{
			for (;;)
			{
                                if (!g_strncasecmp(c, "/A>", 3))
                                        break;

				g_string_append_c (out, *c);
				++c;
                                if (!*c)
                                        break;
                        }
                }

		/* add an anchor to a url */
		else if ((!g_strncasecmp(c, "http://", 7) ||
			 (!g_strncasecmp(c, "https://", 8))))
		{
			const gchar * end = end_of_url (c);
			gchar * url = g_strndup (c, end-c);
			g_string_sprintfa (out,
				"<A HREF=\"%s\">%s</A>", url, url);
			g_free (url);
			c = end;
g_message ("remaining: [%s]", c);
		}

		/* add anchors to likely web sites */
		else if (!g_strncasecmp(c, "www.", 4) &&
		          g_strncasecmp(c, "www..", 5))
		{
			const gchar * end = end_of_url (c);
			if (end-c != 4)
			{
				gchar * url = g_strndup (c, end-c);
				g_string_sprintfa (out,
					"<A HREF=\"http://%s\">%s</A>",
					url, url);
				g_free (url);
				c = end;
			}
                }

		/* add anchors to ftp links */
		else if (!g_strncasecmp(c, "ftp://", 6))
		{
			const gchar * end = end_of_url (c);
			gchar * url = g_strndup (c, end-c);
			g_string_sprintfa (out,
				"<A HREF=\"ftp://%s\">%s</A>", url, url);
			g_free (url);
			c = end;
		}

		/* more ftp anchors */
		else if (!strncasecmp(c, "ftp.", 4))
		{
			const gchar * end = end_of_url (c);
			if (end-c != 4)
			{
				gchar * url = g_strndup (c, end-c);
				g_string_sprintfa (out,
					"<A HREF=\"ftp://%s\">%s</A>",
					url, url);
				g_free (url);
				c = end;
                        }
                }

		/* anchor a mailto */
		else if (!g_strncasecmp(c, "mailto:", 7))
		{
			const gchar * end = end_of_url(c);
			gchar * url = g_strndup (c, end-c);
			g_string_sprintfa (out,
				"<a href=\"%s\">%s</a>",
				url, url);
			g_free (url);
			c = end;
                }

#if 0
		/* possible mailto */
		else if (!strncasecmp(c, "@", 1))
		{
			const gchar * t = c;
			const gchar * end = end_of_url (c);
			gchar * url;
			while (!badchar(*t) && t!=text) --t;
			if (*t=='<') ++t;
			url = g_strndup (t, end-t);
			if (GNKSA_OK == gnksa_check_address (url))
			{
				g_string_sprintfa (out,
					"<a href=\"mailto:%s\">%s</a>",
					url, url);
			}
			g_free (url);
			c = end;
                }
#endif

		/* keep appending */
		else
		{
			g_string_append_c (out, *c++);
		}
        }

g_message ("returning [%s]", out->str);
	c = out->str;
	g_string_free (out, FALSE);
	return (gchar*)c;
}

/****
*****
****/


#if 0
/*
 * for converting a dotted quad to a valid ip, e.g. 192.168.1.2
 */
guint32
gchar_to_ip (const char *str)
{
	struct in_addr ia;
	gint r;
	r = inet_aton (str, &ia);
	if (r)
		return g_ntohl (ia.s_addr);
	else
		return 0;
}

/*
 * for converting a hostname to a valid ip, e.g. www.superpimp.org
 */
guint32
host_to_ip (const char *host)
{
	struct hostent *he;

	g_return_val_if_fail (host!=NULL, 0);

	he = gethostbyname (host);
	if (he)
		return g_ntohl (*(guint32 *) (he->h_addr_list[0]));
	else
		g_warning ("gethostbyname() failed");

	return 0;
}
#endif


/****
*****  DATE
****/

/* converts struct tm to time_t, but does not take the local
   timezone into account unless ``local'' is nonzero.
   This is lifted from mutt 1.2.4, which is GPL and
   copyrighted by Michael R. Elkins <me@cs.hmc.edu> */
static time_t
mutt_mktime (struct tm *t)
{
	time_t g;

	static int AccumDaysPerMonth[12] = {
		0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334
	};

	/* Compute the number of days since January 1 in the same year */
	g = AccumDaysPerMonth [t->tm_mon % 12];

	/* The leap years are 1972 and every 4. year until 2096,
	 * but this algoritm will fail after year 2099 */
	g += t->tm_mday;
	if ((t->tm_year % 4) || t->tm_mon < 2)
		g--;
	t->tm_yday = g;

	/* Compute the number of days since January 1, 1970 */
	g += (t->tm_year - 70) * 365;
	g += (t->tm_year - 69) / 4;

	/* Compute the number of hours */
	g *= 24;
	g += t->tm_hour;

	/* Compute the number of minutes */
	g *= 60;
	g += t->tm_min;

	/* Compute the number of seconds */
	g *= 60;
	g += t->tm_sec;

	return g;
}

gchar*
rfc822_date_generate (time_t secs)
{
	static const gchar *const months[] = {
		"Jan","Feb","Mar","Apr","May","Jun",
		"Jul","Aug","Sep","Oct","Nov","Dec"
	};
	static const gchar *const days[] = {
		"Sun","Mon","Tue","Wed","Thu","Fri","Sat"
	};

	int diff;
	int tz_diff_hours;
	int tz_diff_minutes;
	gchar fmt[64] = { '\0' };
	gchar buf[64] = { '\0' };
	struct tm ut;
	struct tm local;

	/* if 0 passed in, use current time */
	if (secs == 0)
		secs = time (NULL);

	/* figure out the difference between this timezone and UT */
	localtime_r (&secs, &local);
	gmtime_r (&secs, &ut);
	diff = abs ((int) difftime (mutt_mktime(&local), mutt_mktime(&ut)));
	tz_diff_hours = diff / 3600;
	diff %= 3600;
	tz_diff_minutes = diff / 60;

	/* decide on the time string format.  see son of 1036 sec 5.1 for
	 * the rationale of the offset-from-ut notation.
         *
	 * Update: INN 1.7.2-4 (and others) don't accept the <TZ> string at
	 * the end of son-of-1036's date string.  Thanks to Valentini Jove
	 * for testing to see what worked & what didn't.
	 *
	 * Update 2: Big Thanks to tov are jacobsen for investigating what
	 * format other programs use, for suggesting code, and for noting
	 * that we need to roll our own day-of-week and month abbreviations
	 * because son-of-1036 requires English text but strftime() localizes
	 * (Okt instead of Oct or Man instead of Mon).
	 *
	 * Update 3: Gaelyne Gasson spotted a bug that the minute tz offset
	 * should have been divided by 60 to give minutes instead of seconds.
	 * This was causing problems in that tz, which is +0930.  I'm feeling
	 * pretty dumb here... :)
	 */
	sprintf (fmt, "%s, %%d %s %%Y %%H:%%M:%%S %c%02d%02d",
		days[local.tm_wday],
		months[local.tm_mon],
		(diff<0?'-':'+'),
		tz_diff_hours,
		tz_diff_minutes);

	/* generate the time */
	strftime (buf, sizeof(buf), fmt, &local);
	return g_strdup (buf);
}
